#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import contextlib
import datetime as dtm
import inspect
import itertools
import re
from pathlib import Path
from typing import NamedTuple, override

from .._changenoteinfo import ChangeNoteInfo
from .._utils.filename import FileName
from .._utils.types import PathLike, VUIDInput
from .._version import Version
from ..abc import VersionScanner
from ..error import ChanGoError, ValidationError
from ..helpers import ensure_uid

_DEFAULT_PATTERN = re.compile(r"(?P<uid>[^_]+)_(?P<date>[\d-]+)")


class _VersionInfo(NamedTuple):
    date: dtm.date
    directory: Path


class _FileInfo(NamedTuple):
    uid: str
    file: Path


def _make_relative_to(base: Path, path: Path) -> Path:
    if path.is_absolute():
        return path.resolve().absolute()
    return (base / path).resolve().absolute()


class DirectoryVersionScanner(VersionScanner):
    """Implementation of a version scanner that assumes that change notes are stored in
    subdirectories named after the version identifier.

    Args:
        base_directory (:obj:`str` | :class:`~pathlib.Path`): The base directory to scan for
            version directories.

            Important:
                If the path is relative, it will be resolved relative to the
                directory of the calling module.

                .. admonition:: Example

                    If you build your :class:`DirectoryVersionScanner` within
                    ``/home/user/project/chango.py``,
                    passing ``base_directory="changes"`` will resolve to
                    ``/home/user/project/changes``.
        unreleased_directory (:obj:`str` | :class:`~pathlib.Path`): The directory that contains
            unreleased changes.

            Important:
                If :meth:`pathlib.Path.is_dir` returns :obj:`False` for this
                directory, it will be assumed to be a subdirectory of the
                :paramref:`base_directory`.
        directory_pattern (:obj:`str` | :obj:`re.Pattern`, optional): The pattern to match version
            directories against. Must contain one named group ``uid`` for the version identifier
            and a second named group for the ``date`` for the date of the version release in ISO
            format.

    Attributes:
        base_directory (:class:`~pathlib.Path`): The base directory to scan for version
            directories.
        directory_pattern (:obj:`re.Pattern`): The pattern to match version directories against.
        unreleased_directory (:class:`~pathlib.Path`): The directory that contains unreleased
            changes.

    """

    def __init__(
        self,
        base_directory: PathLike,
        unreleased_directory: PathLike,
        directory_pattern: str | re.Pattern[str] = _DEFAULT_PATTERN,
    ):
        self.directory_pattern: re.Pattern[str] = re.compile(directory_pattern)

        caller_dir = Path(inspect.stack()[1].filename).resolve().absolute().parent
        self.base_directory: Path = _make_relative_to(caller_dir, Path(base_directory))
        if not self.base_directory.is_dir():
            raise ValueError(f"Base directory '{self.base_directory}' does not exist.")

        if (path := Path(unreleased_directory)).is_dir():
            self.unreleased_directory: Path = path.resolve().absolute()
        else:
            self.unreleased_directory = self.base_directory / unreleased_directory
            if not self.unreleased_directory.is_dir():
                raise ValueError(
                    f"Unreleased directory '{self.unreleased_directory}' does not exist."
                )

        self.__available_versions: dict[str, _VersionInfo] | None = None

    @property
    def _available_versions(self) -> dict[str, _VersionInfo]:
        # Simple Cache for the available versions
        if self.__available_versions is not None:
            return self.__available_versions

        self.__available_versions = {}
        for directory in self.base_directory.iterdir():
            if not directory.is_dir() or not (
                match := self.directory_pattern.match(directory.name)
            ):
                continue

            uid = match.group("uid")
            date = dtm.date.fromisoformat(match.group("date"))
            self.__available_versions[uid] = _VersionInfo(date, directory)

        return self.__available_versions

    def _get_available_version(self, uid: str) -> Version:
        try:
            return Version(uid=uid, date=self._available_versions[uid].date)
        except KeyError as exc:
            raise ChanGoError(f"Version '{uid}' not available.") from exc

    def invalidate_caches(self) -> None:
        self.__available_versions = None

    @override
    def is_available(self, uid: VUIDInput) -> bool:
        if uid is None:
            return self.has_unreleased_changes()

        if (version := self._available_versions.get(ensure_uid(uid))) is None:
            return False

        if isinstance(uid, Version):
            return version.date == uid.date

        return True

    @override
    def has_unreleased_changes(self) -> bool:
        """Implementation of :meth:`chango.abc.VersionScanner.has_unreleased_changes`.
        Checks if :attr:`unreleased_directory` contains any files.

        Returns:
            :obj:`bool`: :obj:`True` if there are unreleased changes, :obj:`False` otherwise.
        """
        return bool(self._get_file_names(None))

    @override
    def get_latest_version(self) -> Version:
        """Implementation of :meth:`chango.abc.VersionScanner.get_latest_version`.

        Important:
            In case of multiple releases on the same day,
            lexicographical comparison of the version identifiers is employed.

        Returns:
            :class:`~chango.Version`: The latest version
        """
        if not self._available_versions:
            raise ChanGoError("No versions available.")
        return self._get_available_version(
            max(
                self._available_versions, key=lambda uid: (self._available_versions[uid].date, uid)
            )
        )

    @override
    def get_available_versions(
        self, start_from: VUIDInput = None, end_at: VUIDInput = None
    ) -> tuple[Version, ...]:
        """Implementation of :meth:`chango.abc.VersionScanner.get_available_versions`.

        Important:
            Limiting the version range by
            :paramref:`~chango.abc.VersionScanner.get_available_versions.start_from` and
            :paramref:`~chango.abc.VersionScanner.get_available_versions.end_at` is based on
            lexicographical comparison of the version identifiers.

        Returns:
            Tuple[:class:`~chango.Version`]: The available versions within the specified range.
        """
        start = ensure_uid(start_from)
        end = ensure_uid(end_at)
        return tuple(
            self._get_available_version(uid)
            for uid in self._available_versions
            if (start is None or uid >= start) and (end is None or uid <= end)
        )

    def _get_file_names(self, uid: VUIDInput) -> tuple[_FileInfo, ...]:
        try:
            directory = (
                self._available_versions[ensure_uid(uid)].directory
                if uid
                else self.unreleased_directory
            )
        except KeyError as exc:
            raise ChanGoError(f"Version '{uid}' not available.") from exc

        out = []
        # Sorting is an undocumented implementation detail for now!
        for change in sorted(directory.iterdir()):
            if not change.is_file():
                continue

            with contextlib.suppress(ValidationError):
                name = FileName.from_string(change.name)
                out.append(_FileInfo(name.uid, change))

        return tuple(out)

    @override
    def lookup_change_note(self, uid: str) -> ChangeNoteInfo:
        try:
            version, file_name = next(
                (
                    self._get_available_version(version_uid) if version_uid else None,
                    file_info.file.name,
                )
                for version_uid in itertools.chain(self._available_versions, (None,))
                for file_info in self._get_file_names(version_uid)
                if uid == file_info.uid
            )
            directory = (
                self._available_versions[version.uid].directory
                if version
                else self.unreleased_directory
            )
        except StopIteration as exc:
            raise ChanGoError(f"Change note '{uid}' not found in any version.") from exc

        return ChangeNoteInfo(uid, version, directory / file_name)

    @override
    def get_version(self, uid: str) -> Version:
        return self._get_available_version(uid)

    @override
    def get_changes(self, uid: VUIDInput) -> tuple[str, ...]:
        return tuple(file_info.uid for file_info in self._get_file_names(uid))



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
def indent_multiline(text: str, indent: int = 2, newlines: int = 1) -> str:
    """Indent all lines of a multi-line string except the first one."""
    return (newlines * "\n").join(
        line if i == 0 else " " * indent + line for i, line in enumerate(text.splitlines())
    )



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, override

from .._utils.types import VUIDInput
from ..abc import ChangeNote, ChanGo, VersionHistory, VersionNote
from ..action import ChanGoActionData
from ..error import ChanGoError
from ._directoryversionscanner import DirectoryVersionScanner
from .sections import SectionChangeNote, SectionVersionNote

if TYPE_CHECKING:
    from chango import Version


class DirectoryChanGo[VHT: VersionHistory, VNT: VersionNote, CNT: ChangeNote](
    ChanGo[DirectoryVersionScanner, VHT, VNT, CNT]
):
    """Implementation of the :class:`~chango.abc.ChanGo` interface that works with
    :class:`~chango.concrete.DirectoryVersionScanner` and assumes that change notes are stored in
    subdirectories named after the version identifier.

    Args:
        change_note_type (:class:`type`): The type of change notes to load. Must be a subclass of
            :class:`~chango.abc.ChangeNote`.
        version_note_type (:class:`type`): The type of version notes to load. Must be a subclass of
            :class:`~chango.abc.VersionNote`.
        version_history_type (:class:`type`): The type of version histories to load. Must be a
            subclass of :class:`~chango.abc.VersionHistory`.
        scanner (:class:`~chango.concrete.DirectoryVersionScanner`): The version scanner to use.
        directory_format (:obj:`str`, optional): Reverse of
            :paramref:`~chango.concrete.DirectoryVersionScannerdirectory_pattern`.
            Must be a string that can be used
            with :meth:`str.format` and contain at least one named field ``uid`` for the version
            identifier and optionally a second named field ``date`` for the date of the version
            release in ISO format. The default value is compatible with the default value of
            :paramref:`~chango.concrete.DirectoryVersionScannerdirectory_pattern`.

    Attributes:
        directory_format (:obj:`str`): The format string used to create version directories.
    """

    def __init__(
        self: "DirectoryChanGo[VHT, VNT, CNT]",
        change_note_type: type[CNT],
        version_note_type: type[VNT],
        version_history_type: type[VHT],
        scanner: DirectoryVersionScanner,
        directory_format: str = "{uid}_{date}",
    ):
        self._scanner: DirectoryVersionScanner = scanner
        self.directory_format: str = directory_format
        self.change_note_type: type[CNT] = change_note_type
        self.version_note_type: type[VNT] = version_note_type
        self.version_history_type: type[VHT] = version_history_type

    @property
    @override
    def scanner(self) -> DirectoryVersionScanner:
        return self._scanner

    @override
    def build_template_change_note(self, slug: str, uid: str | None = None) -> CNT:
        return self.change_note_type.build_template(slug=slug, uid=uid)

    @override
    def build_version_note(self, version: Optional["Version"]) -> VNT:
        """Implementation of :meth:`~chango.abc.ChanGo.build_version_note`.
        Includes special handling for :class:`~chango.concrete.sections.SectionVersionNote`, which
        has the required argument
        :paramref:`~chango.concrete.sections.SectionVersionNote.section_change_note_type`.
        """
        if issubclass(self.version_note_type, SectionVersionNote):
            return self.version_note_type(
                section_change_note_type=self.change_note_type, version=version
            )
        return self.version_note_type(version=version)

    @override
    def build_version_history(self) -> VHT:
        return self.version_history_type()

    @override
    def load_change_note(self, uid: str) -> CNT:
        return self.change_note_type.from_file(self.scanner.lookup_change_note(uid).file_path)

    @override
    def get_write_directory(self, change_note: CNT | str, version: VUIDInput) -> Path:
        if version is None:
            directory = self.scanner.unreleased_directory
        else:
            if isinstance(version, str):
                try:
                    version_obj = self.scanner.get_version(version)
                except ChanGoError as exc:
                    raise ChanGoError(
                        f"Version '{version}' not available yet. To get the write directory for a "
                        "new version, pass the version as `change.Version` object."
                    ) from exc
            else:
                version_obj = version

            directory = self.scanner.base_directory / self.directory_format.format(
                uid=version_obj.uid, date=version_obj.date.isoformat()
            )

        directory.mkdir(parents=True, exist_ok=True)
        return directory

    @override
    def build_github_event_change_note(
        self, event: dict[str, Any], data: dict[str, Any] | ChanGoActionData | None = None
    ) -> CNT | None:
        """Implementation of :meth:`~chango.abc.ChanGo.build_github_event_change_note`.

        Important:
            By default, this will always call :meth:`chango.abc.ChangeNote.build_from_github_event`
            and does not check if a new change note is necessary.
            The only exception is when :paramref:`~DirectoryChanGo.change_note_type` is a subclass
            of :class:`~chango.concrete.sections.SectionChangeNote`:

            * If there already is a change note for the current pull request, it is updated with
              the new information. If nothing changed, returns :obj:`None`.
            * If the ``data`` parameter is
              an instance of :class:`~chango.action.ChanGoActionData` with a parent pull request,
              then this method will try to find an existing *unreleased* change note for the
              parent pull request and append the new information to it.
        """
        change_note = self.change_note_type.build_from_github_event(event=event, data=data)

        if not isinstance(change_note, SectionChangeNote):
            return change_note

        # Special handling for SectionChangeNote
        orig_data = change_note.to_bytes()
        was_modified = False
        existing_change_notes = self.load_version_note(None).values()

        # First check if we can override any existing change notes
        pr_ids = [pr.uid for pr in change_note.pull_requests]
        for existing_change_note in existing_change_notes:
            if not any(
                uid in pr_ids for uid in (pr.uid for pr in existing_change_note.pull_requests)
            ):
                continue

            change_note.update_uid(existing_change_note.uid)
            orig_data = existing_change_note.to_bytes()
            was_modified = True

        # Handle Parent PRs
        if not isinstance(data, ChanGoActionData) or not data.parent_pull_request:
            if was_modified and (orig_data == change_note.to_bytes()):
                return None
            return change_note

        parent_pr = data.parent_pull_request

        # load all unreleased change notes and find the one for the parent pull request
        for existing_change_note in existing_change_notes:
            if str(parent_pr.number) not in (pr.uid for pr in existing_change_note.pull_requests):
                continue

            # Combine the PRs on existing and new change notes. Override with new PRs if necessary
            existing_prs = {pr.uid: pr for pr in existing_change_note.pull_requests}
            existing_prs.update({pr.uid: pr for pr in change_note.pull_requests})
            existing_change_note.pull_requests = tuple(existing_prs.values())

            for section_name in change_note.SECTIONS:
                if not (new_value := getattr(change_note, section_name)):
                    continue
                if not (existing_value := getattr(existing_change_note, section_name)):
                    setattr(existing_change_note, section_name, new_value)
                else:
                    setattr(existing_change_note, section_name, f"{existing_value}\n{new_value}")

            # change_note must have at least one section specified, so we now that here we have
            # at least one change that we report
            return existing_change_note

        # If we get here, then we didn't find an existing change note for the parent PR
        # This we need to return the new change note
        return change_note



import os
import re
import sys
import tomllib
import typing
from pathlib import Path

import click
from docutils.nodes import Node, reference
from sphinx.addnodes import pending_xref
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment

sys.path.insert(0, str(Path("../..").resolve().absolute()))

from chango import __version__

pyproject_toml = tomllib.loads(Path("../../pyproject.toml").read_text())

project = pyproject_toml["project"]["name"]
version = __version__
release = __version__
documentation_summary = pyproject_toml["project"]["description"]
author = pyproject_toml["project"]["authors"][0]["name"]
copyright = "2024, Hinrich Mahler"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_click",
    "sphinx_copybutton",
    "sphinx_paramlinks",
    "chango.sphinx_ext",
]

html_theme = "furo"

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pydantic": ("https://docs.pydantic.dev/latest/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}

nitpicky = True

# paramlinks options
paramlinks_hyperlink_param = "name"

# Use "Example:" instead of ".. admonition:: Example"
napoleon_use_admonition_for_examples = True

# Don't copy the ">>>" part of interactive python examples
copybutton_only_copy_prompt_lines = False

# Configuration for the chango sphinx directive
chango_pyproject_toml_path = Path(__file__).parent.parent.parent

# Don't show type hints in the signature - that just makes it hardly readable
# and we document the types anyway
autodoc_typehints = "none"
autodoc_member_order = "alphabetical"
autodoc_inherit_docstrings = False

html_static_path = ["../../logo"]

# Theme options are theme-specific and customize the look and feel of a theme
# further. For a list of options available for each theme, see the documentation.
html_theme_options = {
    "light_logo": "chango_light_mode_1024.png",
    "dark_logo": "chango_dark_mode_1024.png",
    "navigation_with_keys": True,
    "footer_icons": [
        {  # Github logo
            "name": "GitHub",
            "url": "https://github.com/Bibo-Joshi/chango",
            "html": (
                '<svg stroke="currentColor" fill="currentColor" stroke-width="0" '
                'viewBox="0 0 16 16"><path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 '
                "2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.4"
                "9-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23"
                ".82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 "
                "0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.2"
                "7 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.5"
                "1.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 "
                '1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z">'
                "</path></svg>"
            ),
            "class": "",
        }
    ],
}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> <release> documentation".
html_title = f"chango {version}"

# Furo's default permalink icon is `#` which doesn't look great imo.
html_permalinks_icon = "¶"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "../../logo/chango_icon.ico"


# Due to Sphinx behaviour, these imports only work when imported here, not at top of module.
from docs.auxil.rich_to_rst import RichConverter  # noqa: E402


def convert_rich_to_rst(_: Sphinx, __: click.Context, lines: list[str]) -> None:
    converter = RichConverter("\n".join(lines))
    converter.parse_rich_text()
    lines[:] = converter.render_rst_text().split("\n")


_TYPE_VAR_PATTERN = re.compile(r"typing\.([A-Z][a-zA-Z]*)")


def missing_reference(
    _: Sphinx, __: BuildEnvironment, node: pending_xref, contnode: Node
) -> None | Node:
    """Here we redirect links to type variables. Sphinx tries to link TypeVar T to typing.T which
    does not exist. We instead link to typing.TypeVar.
    """
    if not (match := _TYPE_VAR_PATTERN.match(node["reftarget"])):
        # Sort out everything that is obviously not a TypeVar
        return None

    name = match.group(1)
    if hasattr(typing, name):
        # We don't want to change valid links that exist
        return None

    link_node = reference(refuri="https://docs.python.org/3/library/typing.html#typing.TypeVar")
    link_node.append(contnode.deepcopy())
    return link_node


def config_inited(_: Sphinx, __: dict[str, str]) -> None:
    # for usage in _cli.__init__
    os.environ["SPHINX_BUILD"] = "True"


def setup(app: Sphinx) -> None:
    app.connect("config-inited", config_inited)
    app.connect("missing-reference", missing_reference)

    # We use the hooks defined by sphinx-click to convert python-rich syntax to rst
    # See also https://sphinx-click.readthedocs.io/en/latest/usage/#events.
    # This is a relatively sane way to have nice rendering both in the CLI and in the HTML
    # documentation. Doing conversion for computing the documentation is preferable to doing
    # conversions for the CLI as the latter would impact the CLI performance
    for event in [
        "sphinx-click-process-description",
        "sphinx-click-process-usage",
        "sphinx-click-process-options",
        "sphinx-click-process-arguments",
        "sphinx-click-process-envvars",
        "sphinx-click-process-epilog",
    ]:
        app.connect(event, convert_rich_to_rst)



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import pytest
import shortuuid

from chango.concrete.sections import (
    GitHubSectionChangeNote,
    PullRequest,
    Section,
    SectionChangeNote,
)
from chango.constants import MarkupLanguage
from chango.error import ValidationError


class DummyChangNote(
    GitHubSectionChangeNote.with_sections(
        [
            Section(uid="req_section", title="Required Section", is_required=True),
            Section(uid="opt_section", title="Optional Section"),
        ]
    )
):
    OWNER = "my-username"
    REPOSITORY = "my-repo"


@pytest.fixture
def section_change_note():
    return DummyChangNote(
        slug="slug",
        req_section="req 𝛙𝌢𑁍",
        opt_section="opt 𝛙𝌢𑁍",
        pull_requests=(
            PullRequest(uid="uid1", closes_threads=("thread1",), author_uids=("author1",)),
            PullRequest(uid="uid2", closes_threads=("thread2",), author_uids=("author2",)),
        ),
    )


class TestSectionChangeNote:
    """Since TestSectionChangeNote is an abstract base class, we are testing with
    GitHubTestSectionChangeNote as a simple implementation.

    Note that we do *not* test abstract methods, as that is the responsibility of the concrete
    implementations.
    """

    sections = (
        Section(uid="req_section", title="Required Section", is_required=True),
        Section(uid="opt_section", title="Optional Section"),
    )

    def test_manual_subclass(self):
        class SubClass(SectionChangeNote):
            @classmethod
            def get_pull_request_url(cls, pr_uid: str) -> str:
                pass

            @classmethod
            def get_thread_url(cls, thread_uid: str) -> str:
                pass

            @classmethod
            def get_author_url(cls, author_uid: str) -> str:
                pass

        with pytest.raises(TypeError, match="SectionChangeNote must not be subclassed manually"):
            SubClass(slug="slug")

    @pytest.mark.parametrize("name", [None, "CustomName"])
    def test_with_sections(self, name):
        cls = SectionChangeNote.with_sections(self.sections, name=name)
        assert cls.SECTIONS["req_section"] is self.sections[0]
        assert cls.SECTIONS["opt_section"] is self.sections[1]
        assert cls.__name__ == (name or "DynamicSectionChangeNote")

    def test_with_sections_empty_sequence(self):
        with pytest.raises(ValueError, match="Class must have at least one section"):
            SectionChangeNote.with_sections([])

    def test_empty_init(self):
        cls = GitHubSectionChangeNote.with_sections(
            [Section(uid=f"opt_{i}", title=f"Optional {i}") for i in range(10)]
        )
        with pytest.raises(ValidationError, match="At least one section must be specified"):
            cls(slug="slug", opt_0="", opt_1=None)

    def test_constants(self, section_change_note):
        assert section_change_note.file_extension == "toml"
        assert section_change_note.MARKUP == MarkupLanguage.RESTRUCTUREDTEXT

    @pytest.mark.parametrize("has_prs", [True, False])
    def test_from_string(self, section_change_note, has_prs):
        string = """
req_section = '''Required section.
With multiple lines.'''

opt_section = "Optional Section."
"""
        if has_prs:
            string += """
[[pull_requests]]
uid = "uid1"
closes_threads = ["thread1", "thread2"]
author_uids = ["author1"]

[[pull_requests]]
uid = "uid2"
closes_threads = ["thread3"]
author_uid = ["author2"]
"""

        change_note = section_change_note.from_string("slug", "uid", string)
        assert change_note.slug == "slug"
        assert change_note.uid == "uid"
        assert change_note.req_section == "Required section.\nWith multiple lines."
        assert change_note.opt_section == "Optional Section."
        if has_prs:
            assert len(change_note.pull_requests) == 2  # noqa: PLR2004
            assert change_note.pull_requests[0].uid == "uid1"
            assert change_note.pull_requests[0].closes_threads == ("thread1", "thread2")
            assert change_note.pull_requests[0].author_uids == ("author1",)
            assert change_note.pull_requests[1].uid == "uid2"
            assert change_note.pull_requests[1].closes_threads == ("thread3",)
            assert change_note.pull_requests[1].author_uids == ("author2",)
        else:
            assert change_note.pull_requests == ()

    def test_from_string_invalid(self, section_change_note):
        with pytest.raises(ValidationError, match="Invalid TOML data"):
            section_change_note.from_string("slug", "uid", "invalid toml")

    @pytest.mark.parametrize("encoding", ["utf-8", "utf-16"])
    def test_to_string(self, section_change_note, encoding):
        string = section_change_note.to_string(encoding=encoding)
        assert (
            string
            == """req_section = "req 𝛙𝌢𑁍"
opt_section = "opt 𝛙𝌢𑁍"
[[pull_requests]]
uid = "uid1"
author_uids = ["author1"]
closes_threads = ["thread1"]

[[pull_requests]]
uid = "uid2"
author_uids = ["author2"]
closes_threads = ["thread2"]
"""
        )

    @pytest.mark.parametrize("uid", ["uid1", None])
    def test_build_template(self, uid):
        change_note = DummyChangNote.build_template("slug", uid)
        assert change_note.req_section == "Required Section Content"
        assert change_note.opt_section == "Optional Section Content"
        assert change_note.pull_requests == (
            PullRequest(
                uid="pr-number-1", closes_threads=("thread1", "thread2"), author_uids=("author1",)
            ),
            PullRequest(uid="pr-number-2", closes_threads=("thread3",), author_uids=("author2",)),
        )

        assert change_note.slug == "slug"
        if uid:
            assert change_note.uid == "uid1"
        else:
            assert isinstance(change_note.uid, str)
            assert len(change_note.uid) == len(shortuuid.ShortUUID().uuid())



# SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
# SPDX-License-Identifier: MIT
import os
import shutil
from pathlib import Path

import pytest


# To make the tests agnostic of the cwd
@pytest.fixture(autouse=True)
def _change_test_dir(request, monkeypatch):
    monkeypatch.chdir(request.config.rootdir)


def test_build():
    assert os.system("python -m build") == 0
    shutil.rmtree(Path("dist"))



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import sys
from collections.abc import Collection
from contextlib import nullcontext
from pathlib import Path
from string import Template
from types import SimpleNamespace

import pytest

from chango.concrete import BackwardCompatibleChanGo
from chango.config import ChanGoConfig, ChanGoInstanceConfig
from tests.auxil.files import PROJECT_ROOT_PATH, data_path, temporary_chdir


class TestChanGoInstanceConfig:
    chango_instance = ChanGoInstanceConfig(name="name", module="module", package="package")

    def test_init(self, tmp_path):
        # To ensure that there is pyproject.toml in the current directory that would interfere
        # with the test.
        with temporary_chdir(tmp_path):
            config = ChanGoConfig(sys_path=Path("sys_path"), chango_instance=self.chango_instance)
            assert config.sys_path == Path("sys_path")
            assert config.chango_instance is self.chango_instance

    def test_init_required(self, tmp_path):
        # To ensure that there is pyproject.toml in the current directory that would interfere
        # with the test.
        with temporary_chdir(tmp_path):
            config = ChanGoConfig(chango_instance=self.chango_instance)
            assert config.sys_path is None
            assert config.chango_instance is self.chango_instance

    @pytest.mark.parametrize(
        "path",
        [
            None,
            data_path("config/pyproject.toml"),
            data_path("config/pyproject.toml").as_posix(),
            data_path("config/pyproject.toml").relative_to(Path.cwd(), walk_up=True),
            data_path("config/pyproject.toml").relative_to(Path.cwd(), walk_up=True).as_posix(),
            data_path("config"),
            data_path("config").as_posix(),
        ],
        ids=[
            "None",
            "absolute",
            "absolute-string",
            "relative",
            "relative-string",
            "directory",
            "directory-string",
        ],
    )
    def test_load_path_input(self, path):
        with temporary_chdir(data_path("config")) if path is None else nullcontext():
            config = ChanGoConfig.load(path)
            assert config.sys_path == Path("/abs/sys_path").absolute()
            assert config.chango_instance == self.chango_instance

    @pytest.mark.parametrize(
        ("sys_path", "expected"),
        [
            (None, None),
            (Path("/abs/sys_path").absolute(), Path("/abs/sys_path").absolute()),
            (Path("relative"), data_path("config/relative").absolute()),
            (Path("../relative"), data_path("relative").absolute()),
        ],
        ids=["None", "absolute", "relative", "relative-parent"],
    )
    def test_load_sys_path_output(self, sys_path, expected):
        tmp_file = data_path("config/tmp.toml")
        try:
            sys_path_entry = f"sys_path = '{sys_path}'" if sys_path else ""
            template = Template(
                data_path("config/pyproject.toml.template").read_text(encoding="utf-8")
            )
            tmp_file.write_text(
                template.substitute(sys_path_entry=sys_path_entry), encoding="utf-8"
            )

            config = ChanGoConfig.load(tmp_file)
            assert config.sys_path == expected
        finally:
            tmp_file.unlink()

    def test_import_chango_instance_basic(self):
        # Testing with importlib is a bit of a hassle, and also we don't want to test
        # the actual import, but the import logic. So here we just test that the
        # config of the chango repo itself is imported correctly.
        with temporary_chdir(PROJECT_ROOT_PATH):
            config = ChanGoConfig.load()
            chango_instance = config.import_chango_instance()
            assert isinstance(chango_instance, BackwardCompatibleChanGo)

    @pytest.mark.parametrize(
        ("sys_path", "expected"),
        [
            (None, None),
            (Path("/abs/sys_path").absolute(), Path("/abs/sys_path").absolute()),
            (Path("relative"), Path.cwd() / "relative"),
        ],
        ids=["None", "relative", "relative-parent"],
    )
    def test_import_chango_instance_sys_path_input(self, monkeypatch, sys_path, expected):
        original_sys_path = sys.path.copy()

        # Here we test that the sys_path is added to the system path correctly.
        def is_in_sys_path(system_path: Collection[str], search_path: Path):
            sys_paths = map(Path, system_path)
            return any(search_path == path for path in sys_paths)

        def import_module(*_, **__):
            return SimpleNamespace(name=sys.path.copy())

        monkeypatch.setattr("importlib.import_module", import_module)

        reported_path = ChanGoConfig(
            sys_path=sys_path, chango_instance=self.chango_instance
        ).import_chango_instance()

        if expected is not None:
            assert is_in_sys_path(reported_path, expected)
        else:
            assert reported_path == original_sys_path

        # Check that everything was restored correctly
        assert sys.path == original_sys_path



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

from pathlib import Path

import pytest
from click import UsageError

from chango.constants import MarkupLanguage
from tests.cli.conftest import ReuseCliRunner


class TestReport:
    @pytest.mark.parametrize(
        "markup",
        [MarkupLanguage.MARKDOWN, MarkupLanguage.HTML, None],
        ids=["Markdown", "HTML", "DefaultMarkup"],
    )
    @pytest.mark.parametrize(
        "markup_option", ["--markup", "-m"], ids=["LongMarkup", "ShortMarkup"]
    )
    @pytest.mark.parametrize(
        ("output", "output_option"), [(None, None), (True, "--output"), (True, "-o")]
    )
    def test_report_version(
        self,
        cli: ReuseCliRunner,
        mock_chango_instance,
        markup,
        markup_option,
        output,
        output_option,
        tmp_path: Path,
    ):
        file_path = tmp_path / "output_file"
        version_note = mock_chango_instance.load_version_note.return_value
        version_note.render.return_value = "expected_render_output"

        args = ["report", "version", "--uid", "1.2.3"]
        if markup is not None:
            args.extend([markup_option, markup.value])
        if output is True:
            args.extend([output_option, file_path.as_posix()])

        result = cli.invoke(args=args)

        assert result.check_exit_code()

        if output is not None:
            assert result.stdout == f"Report written to {file_path}\n"
        else:
            assert result.stdout == "expected_render_output\n"

        mock_chango_instance.load_version_note.assert_called_once_with("1.2.3")
        version_note.render.assert_called_once_with(markup=markup or MarkupLanguage.MARKDOWN)
        if output is True:
            assert file_path.read_text() == "expected_render_output"

    @pytest.mark.parametrize(
        "markup",
        [MarkupLanguage.MARKDOWN, MarkupLanguage.HTML, None],
        ids=["Markdown", "HTML", "DefaultMarkup"],
    )
    @pytest.mark.parametrize(
        "markup_option", ["--markup", "-m"], ids=["LongMarkup", "ShortMarkup"]
    )
    @pytest.mark.parametrize(
        ("output", "output_option"), [(None, None), (True, "--output"), (True, "-o")]
    )
    def test_report_history(
        self,
        cli: ReuseCliRunner,
        mock_chango_instance,
        markup,
        markup_option,
        output,
        output_option,
        tmp_path: Path,
    ):
        file_path = tmp_path / "output_file"
        version_history = mock_chango_instance.load_version_history.return_value
        version_history.render.return_value = "expected_render_output"

        args = ["report", "history"]
        if markup is not None:
            args.extend([markup_option, markup.value])
        if output is True:
            args.extend([output_option, file_path.as_posix()])

        result = cli.invoke(args=args)

        assert result.check_exit_code()

        if output is not None:
            assert result.stdout == f"Report written to {file_path}\n"
        else:
            assert result.stdout == "expected_render_output\n"

        mock_chango_instance.load_version_history.assert_called_once_with()
        version_history.render.assert_called_once_with(markup=markup or MarkupLanguage.MARKDOWN)
        if output is True:
            assert file_path.read_text() == "expected_render_output"

    @pytest.mark.parametrize("invalidity_type", ["dir", "non_writable"])
    @pytest.mark.parametrize("subcommand", ["version", "history"], ids=["Version", "History"])
    def test_report_invalid_file(
        self,
        cli: ReuseCliRunner,
        mock_chango_instance,
        tmp_path: Path,
        invalidity_type,
        subcommand,
    ):
        file_path = tmp_path / "output_file"
        if invalidity_type == "dir":
            output_path = tmp_path
        elif invalidity_type == "non_writable":
            file_path.touch()
            file_path.chmod(0o444)
            output_path = file_path

        args = ["report", subcommand, "--uid", "1.2.3", "--output", output_path.as_posix()]
        result = cli.invoke(args=args)

        assert result.check_exit_code(UsageError.exit_code)
        mock_chango_instance.load_version_note.assert_not_called()

    @pytest.mark.parametrize("subcommand", ["version", "history"], ids=["Version", "History"])
    def test_report_invalid_markup(self, cli: ReuseCliRunner, mock_chango_instance, subcommand):
        args = ["report", subcommand, "--markup", "invalid_markup"]
        if subcommand == "version":
            args.extend(["--uid", "1.2.3"])
        result = cli.invoke(args=args)

        assert result.check_exit_code(UsageError.exit_code)
        mock_chango_instance.load_version_note.assert_not_called()



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
"""This module contains implementations of the interface classes defined in the
:mod:`~chango.abc` module that are shipped with this package."""

__all__ = [
    "BackwardCompatibleChanGo",
    "BackwardCompatibleVersionScanner",
    "CommentChangeNote",
    "CommentVersionNote",
    "DirectoryChanGo",
    "DirectoryVersionScanner",
    "HeaderVersionHistory",
    "sections",
]

from . import sections
from ._backwardcompatiblechango import BackwardCompatibleChanGo
from ._backwardcompatibleversionscanner import BackwardCompatibleVersionScanner
from ._commentchangenote import CommentChangeNote
from ._commentversionnote import CommentVersionNote
from ._directorychango import DirectoryChanGo
from ._directoryversionscanner import DirectoryVersionScanner
from ._headerversionhistory import HeaderVersionHistory



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from pathlib import Path
from typing import TYPE_CHECKING, Union

from .._version import Version

if TYPE_CHECKING:
    from ..abc._changenote import ChangeNote

VersionUID = str | None
VUIDInput = Version | str | None
CNUIDInput = Union["ChangeNote", str]
PathLike = str | Path



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import importlib
import logging
from pathlib import Path
from types import ModuleType

from jinja2 import Template

logger = logging.getLogger(__name__)


CLASS_TEMPLATE = Template(
    """{{ class_name }}
{{ "=" * class_name|length }}

.. autoclass:: {{ fqn }}
    :members:
    :show-inheritance:
"""
)

MODULE_TEMPLATE = Template(
    """{{ module_name }}
{{ "=" * module_name|length }}

.. automodule:: {{ fqn }}

.. toctree::
    :titlesonly:

    {% for child in children -%}
    {{ child }}
    {% endfor -%}
"""
)


def write_text(file_path: Path, content: str) -> None:
    if file_path.exists():
        logger.debug("File %s already exists, skipping", file_path)
        return
    file_path.write_text(content, encoding="utf-8")


def write_class_rst_file(class_name: str, fqn: str, output_dir: Path) -> Path:
    file_name = f"{fqn.lower()}.rst"
    file_path = output_dir / file_name
    write_text(file_path, CLASS_TEMPLATE.render(class_name=class_name, fqn=fqn))

    return file_path


def write_module_rst_file(
    module_name: str, fqn: str, children: list[Path], output_dir: Path
) -> Path:
    file_name = f"{fqn.lower()}.rst"
    file_path = output_dir / file_name
    write_text(
        file_path,
        MODULE_TEMPLATE.render(
            module_name=module_name, fqn=fqn, children=sorted(child.stem for child in children)
        ),
    )

    return file_path


def create_rst_files(module_name: str, base_path: Path) -> list[Path]:
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        logger.error("Could not import module %s", module_name)
        return []

    if not hasattr(module, "__all__"):
        return []

    created_files: list[Path] = []

    for name in module.__all__:
        fqn = f"{module_name}.{name}"
        try:
            member = getattr(module, name)
        except AttributeError:
            continue

        if not isinstance(member, type | ModuleType):
            # Rest is covered on module level
            continue

        if isinstance(member, type):
            created_files.append(write_class_rst_file(name, fqn, base_path))
        else:
            sub_files = create_rst_files(fqn, base_path)
            created_files.append(write_module_rst_file(name, fqn, sub_files, base_path))

    write_module_rst_file(module_name, module_name, created_files, base_path)
    return created_files


def main() -> None:
    output_dir = Path(__file__).parent / "source"
    output_dir.mkdir(exist_ok=True)

    create_rst_files("chango", output_dir)


if __name__ == "__main__":
    main()



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import json
import typing

from docutils.nodes import Node
from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxDirective

from chango._utils.types import PathLike
from chango.config import get_chango_instance
from chango.constants import MarkupLanguage


class JsonValidator:
    """Validator that interprets the input as JSON data and loads it accordingly.
    The value must be a JSON-loadable value, not None.
    """

    def __init__(self, option_name: str) -> None:
        self.option_name = option_name

    def __call__(self, var: str | None) -> str | int | float | bool | dict | list | None:
        if var is None:
            raise ValueError(
                f"Option '{self.option_name}' must be a JSON-loadable value, not None"
            )

        try:
            return json.loads(var)
        except json.JSONDecodeError as exc:
            raise ValueError(
                f"Option '{self.option_name}' must be a JSON-loadable value, not {var!r}"
            ) from exc

    def __repr__(self) -> str:  # pragma: no cover
        """So far only used for debugging."""
        return "<return json parsed data>"


def parse_function(func: typing.Callable) -> dict[str, typing.Callable[[str | None], typing.Any]]:
    """Parse a function's signature and annotations to create a dictionary of validators.
    Custom validators may be defined using the `typing.Annotated` type. Defaults are not -
    options are interpreted as kwargs and are always expected to carry a value.

    Example:
        >>> from typing import Annotated
        >>> from collections.abc import Sequence
        >>>
        >>> def custom_validator(x: str | None) -> Sequence[float]:
        ...     return tuple(map(float, x.split(","))) if x is not None else (1.0, 2.0, 3.0)
        >>>
        >>> def foo(
        >>>     arg1: str,
        >>>     arg2: int = 42,
        >>>     arg3: Annotated[Sequence[float], custom_validator] = (1, 2, 3),
        >>> ) -> None:
        ...     pass
        >>>
        >>> parse_function(foo)
        {'arg1': <return json parsed data>, 'arg2': <return json parsed data>, \
        'arg3': <function main.<locals>.custom_validator at 0x000001FB6F1D7100>}

    """
    # To get custom validator, we need to evaluate the annotations to detect `typing.Annotated`
    annotations = typing.get_type_hints(func, include_extras=True, localns=locals())
    return {
        name: typing.get_args(annotation)[1]
        if typing.get_origin(annotation) is typing.Annotated
        else JsonValidator(name)
        for name, annotation in annotations.items()
        # The return value is not a parameter
        if name != "return"
    }


def directive_factory(app: Sphinx) -> type[SphinxDirective]:
    """Create a directive class that uses the chango instance from the Sphinx app config.
    This approach is necessary because the `option_spec` attribute of a directive class can
    not be dynamically set.
    """
    if not isinstance(app.config.chango_pyproject_toml_path, PathLike | None):
        raise TypeError(
            f"Expected 'chango_pyproject_toml_path' to be a string or Path, "
            f"but got {type(app.config.chango_pyproject_toml_path)}"
        )
    chango_instance = get_chango_instance(app.config.chango_pyproject_toml_path)

    class ChangoDirective(SphinxDirective):
        has_content = True
        option_spec = parse_function(  # type: ignore[assignment]
            chango_instance.load_version_history
        )

        def run(self) -> list[Node]:
            title = " ".join(self.content)
            text = chango_instance.load_version_history(**self.options).render(
                MarkupLanguage.RESTRUCTUREDTEXT
            )
            if title:
                decoration = len(title) * "="
                text = f"{decoration}\n{title}\n{decoration}\n\n{text}"
            return self.parse_text_to_nodes(text, allow_section_headings=True)

    return ChangoDirective



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

from ._cli import app  # pragma: no cover

if __name__ == "__main__":  # pragma: no cover
    app()



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import datetime as dtm
from dataclasses import dataclass


@dataclass(frozen=True)
class Version:
    """Objects of this type represent a released version of a software project.

    Args:
        uid (:obj:`str`): Unique identifier / version number of this version.
        date (:class:`datetime.date`): Release date of this version.

    Attributes:
        uid (:obj:`str`): Unique identifier / version number of this version.
        date (:class:`datetime.date`): Release date of this version.
    """

    uid: str
    date: dtm.date



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from typing import TYPE_CHECKING, override

from .._utils.strings import indent_multiline
from ..abc import VersionNote
from ..concrete import CommentChangeNote
from ..constants import MarkupLanguage
from ..error import UnsupportedMarkupError

if TYPE_CHECKING:
    from chango import Version


class CommentVersionNote[V: (Version, None)](VersionNote[CommentChangeNote, V]):
    """A simple version note implementation that works with
    :class:`~chango.concrete.CommentChangeNote`.
    """

    @override
    def render(self, markup: str) -> str:
        """Render the version note to a string by listing all contained change notes separated
        by a newline.
        For markup languages Markdown, HTML and reStructuredText, the change notes will be
        rendered as unordered lists.

        Args:
            markup (:obj:`str`): The markup language to use for rendering.

        Raises:
            :exc:`~chango.error.UnsupportedMarkupError`: If the ``markup`` parameter does not
                coincide with :attr:`chango.concrete.CommentChangeNote.MARKUP`

        Returns:
            :obj:`str`: The rendered version note.
        """
        try:
            markup = MarkupLanguage.from_string(markup)
        except ValueError as exc:
            raise UnsupportedMarkupError(markup) from exc

        match markup:
            case MarkupLanguage.MARKDOWN:
                return "\n".join(
                    f"- {indent_multiline(note.comment, indent=4, newlines=2)}"
                    for note in self.values()
                )
            case MarkupLanguage.HTML:
                return (
                    "<ul>\n"
                    + "\n".join(
                        f"<li>{note.comment.replace('\n', '<br>')}</li>" for note in self.values()
                    )
                    + "\n</ul>"
                )
            case MarkupLanguage.RESTRUCTUREDTEXT:
                return "\n".join(
                    f"- {indent_multiline(note.comment, newlines=2)}" for note in self.values()
                )
            case _:
                return "\n\n".join(note.comment for note in self.values())



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import pytest

from chango.constants import MarkupLanguage


class TestMarkupLanguage:
    def test_str_instance(self):
        for member in MarkupLanguage:
            assert isinstance(member, str)

    @pytest.mark.parametrize("leading_dot", [True, False])
    def test_from_string_members(self, leading_dot):
        for member in MarkupLanguage:
            string = member.value
            if leading_dot:
                string = "." + string
            assert MarkupLanguage.from_string(string) == member

    @pytest.mark.parametrize("leading_dot", [True, False])
    def test_from_string_default_mapping(self, leading_dot):
        default_mapping = {
            "adoc": MarkupLanguage.ASCIIDOC,
            "htm": MarkupLanguage.HTML,
            "md": MarkupLanguage.MARKDOWN,
            "mkd": MarkupLanguage.MARKDOWN,
            "mdwn": MarkupLanguage.MARKDOWN,
            "mdown": MarkupLanguage.MARKDOWN,
            "mdtxt": MarkupLanguage.MARKDOWN,
            "mdtext": MarkupLanguage.MARKDOWN,
            "mediawiki": MarkupLanguage.MEDIAWIKI,
            "org": MarkupLanguage.ORG,
            "pod": MarkupLanguage.POD,
            "rdoc": MarkupLanguage.RDOC,
            "text": MarkupLanguage.TEXT,
        }
        for ext, member in default_mapping.items():
            string = ext
            if leading_dot:
                string = "." + string
            assert MarkupLanguage.from_string(string) == member

    @pytest.mark.parametrize("leading_dot", [True, False])
    def test_from_string_custom_mapping(self, leading_dot):
        custom_mapping = {
            "abc": MarkupLanguage.TEXT,
            "def": MarkupLanguage.HTML,
            "ghi": MarkupLanguage.MARKDOWN,
        }
        for ext, member in custom_mapping.items():
            string = ext
            if leading_dot:
                string = "." + string
            assert MarkupLanguage.from_string(string, custom_mapping) == member

    @pytest.mark.parametrize("leading_dot", [True, False])
    @pytest.mark.parametrize(
        "mapping", [None, {"abc": MarkupLanguage.TEXT}], ids=["default", "custom"]
    )
    def test_from_string_invalid(self, leading_dot, mapping):
        string = "invalid"
        if leading_dot:
            string = "." + string
        with pytest.raises(ValueError, match=f"File extension `{string}` not found"):
            MarkupLanguage.from_string(string, mapping)



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from dataclasses import dataclass
from pathlib import Path

from ._version import Version


@dataclass(frozen=True)
class ChangeNoteInfo:
    """Objects of this type represents metadata about a change note.

    Args:
        uid (:obj:`str`): Unique identifier of this change note.
        version (:class:`~chango.Version` | :obj:`None`): The version the change note belongs to.
            May be :obj:`None` if the change note is not yet released.
        file_path (:class:`pathlib.Path`): The file path this change note is stored at.

    Attributes:
        uid (:obj:`str`): Unique identifier of this change note.
        version (:class:`~chango.Version` | :obj:`None`): The version the change note belongs to.
            May be :obj:`None` if the change note is not yet released.
        file_path (:class:`pathlib.Path`): The file path this change note is stored at.
    """

    uid: str
    version: Version | None
    file_path: Path



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
"""This module contains an implementation of :class:`~chango.abc.ChangeNote` that consists of
multiple sections and includes references to pull requests that are related to the change.
The main class is :class:`SectionChangeNote`, while :class:`Section` and
:class:`PullRequest` are used to define the sections and pull requests, respectively.

Example:
    To create a change note with two sections, one required and one optional, use

    .. code-block:: python

        from chango.concrete.sections import GitHubSectionChangeNote, Section


        class MySectionChangeNote(
            GitHubSectionChangeNote.with_sections(
                [
                    Section(uid="required_section", title="Required Section", is_required=True),
                    Section(uid="optional_section", title="Optional Section"),
                ]
            )
        ):
            OWNER = "my-username"
            REPOSITORY = "my-repo"
"""

__all__ = [
    "GitHubSectionChangeNote",
    "PullRequest",
    "Section",
    "SectionChangeNote",
    "SectionVersionNote",
]

from ._githubsectionchangenote import GitHubSectionChangeNote
from ._pullrequest import PullRequest
from ._section import Section
from ._sectionchangenote import SectionChangeNote
from ._sectionversionnote import SectionVersionNote



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT


from rich.errors import StyleSyntaxError
from rich.style import Style
from rich.text import Span, Text


class RichConverter:
    """Minimal effort implementation of converting text with python-rich formatting options
    to rST formatted text."""

    def __init__(self, rich_text: str) -> None:
        self.rich_text = rich_text
        self.spans: list[Span] = []
        self.plain_text: str = ""

    def parse_rich_text(self) -> None:
        """Reads the rich-formatted text. Parses it to plain text + a list of formatting entities.
        Those are then stored in :attr:`spans` and :attr:`plain_text`.
        """
        text = Text.from_markup(self.rich_text)

        # Extract plain text
        self.plain_text = text.plain

        # Extract formatting data
        self.spans.extend(text.spans)

    @staticmethod
    def _process_rich_style(plain_text: str, style: Style) -> str:
        """Given a text and a rich Style object, return the corresponding rst string.
        If the style is not known, the plain text is returned unchanged.
        """
        rst_text = plain_text

        if style.italic:
            rst_text = f"*{rst_text}*"
        if style.bold:
            rst_text = f"**{rst_text}**"
        if style.link:
            if not style.link.startswith("https://chango.readthedocs.io/"):
                rst_text = f"`{rst_text} <{style.link}>`_"
            else:
                try:
                    reference = style.link.rsplit("#", 1)[1]
                except IndexError:
                    reference = style.link.rsplit("/", 1)[1].removesuffix(".html")
                rst_text = f":attr:`{reference}`"

        return rst_text

    @staticmethod
    def _process_rich_string(plain_text: str, style: str) -> str:
        """Given a text and a rich style in string representation, return the corresponding rst
        string.
        If the style is not known, the plain text is returned unchanged.
        """
        rst_text = plain_text

        if style == "code":
            rst_text = f"``{rst_text}``"

        return rst_text

    @classmethod
    def process_rich_span(cls, plain_text: str, style: str | Style) -> str:
        """Given a plain text and a rich style in string representation, return the corresponding
        rst formatted string.
        """
        if isinstance(style, str):
            try:
                style_obj: Style | str = Style.parse(style)
            except StyleSyntaxError:
                style_obj = style
        else:
            style_obj = style

        if isinstance(style_obj, Style):
            return cls._process_rich_style(plain_text, style_obj)
        return cls._process_rich_string(plain_text, style_obj)

    @classmethod
    def _render_rst_text(cls, plain_text: str, spans: list[Span], offset: int = 0) -> str:
        # Implementation heavily borrows from
        # https://github.com/python-telegram-bot/python-telegram-bot/blob/…
        # …5ab82a9c2b09286b66777ad0345b1abf3dedf131/telegram/_message.py#L4487-L4574
        rst_text = ""
        last_offset = 0
        sorted_spans = sorted(spans, key=lambda s: s.start)
        parsed_spans = []

        for span in sorted_spans:
            if span in parsed_spans:
                continue

            span_start = span.start - offset
            span_end = span.end - offset
            span_text = plain_text[span_start:span_end]
            nested_spans = [
                s
                for s in sorted_spans
                if s.start >= span_start and s.end <= span_end and s != span
            ]
            parsed_spans.extend(nested_spans)

            if nested_spans:
                span_text = cls._render_rst_text(
                    plain_text=span_text, spans=nested_spans, offset=span_start
                )

            insert = cls.process_rich_span(span_text, span.style)
            rst_text += plain_text[last_offset:span_start] + insert
            last_offset = span_start + (span.end - span.start)

        rst_text += plain_text[last_offset:]
        return rst_text

    def render_rst_text(self) -> str:
        """Render :attr:`plain_text` with rst formatting based on :attr:`spans`."""
        return self._render_rst_text(self.plain_text, self.spans)



# SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
# SPDX-License-Identifier: MIT
__all__ = ["app"]

import os
from typing import Annotated

import typer

from .. import __version__
from .config import app as config_app
from .edit import edit
from .new import new
from .release import release
from .report import app as report_app

app = typer.Typer(
    help="CLI for chango - CHANgelog GOvernor for Your Project", rich_markup_mode="rich"
)


def version_callback(value: bool) -> None:
    if value:
        typer.echo(__version__)
        raise typer.Exit


@app.callback()
def main(
    _version: Annotated[
        bool,
        typer.Option("--version", callback=version_callback, help="Show the version and exit."),
    ] = False,
) -> None:
    pass


app.add_typer(config_app, name="config")
app.command()(edit)
app.command()(new)
app.command()(release)
app.add_typer(report_app, name="report")

if os.getenv("SPHINX_BUILD") == "True":  # pragma: no cover
    # See https://github.com/fastapi/typer/issues/200#issuecomment-795873331
    _typer_click_object = typer.main.get_command(app)



#  noqa: INP001
#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

from chango.concrete import (
    BackwardCompatibleChanGo,
    BackwardCompatibleVersionScanner,
    CommentChangeNote,
    CommentVersionNote,
    DirectoryChanGo,
    DirectoryVersionScanner,
    HeaderVersionHistory,
)
from chango.concrete.sections import GitHubSectionChangeNote, Section, SectionVersionNote
from chango.constants import MarkupLanguage


class RstChangeNote(CommentChangeNote):
    MARKUP = MarkupLanguage.RESTRUCTUREDTEXT


legacy_version_scanner = DirectoryVersionScanner(
    base_directory="./legacy", unreleased_directory="unreleased"
)

legacy_chango_instance = DirectoryChanGo(
    change_note_type=RstChangeNote,
    version_note_type=CommentVersionNote,
    version_history_type=HeaderVersionHistory,
    scanner=legacy_version_scanner,
)

new_version_scanner = DirectoryVersionScanner(
    base_directory=".", unreleased_directory="unreleased"
)
ChangoSectionChangeNote = GitHubSectionChangeNote.with_sections(
    [
        Section(uid="highlights", title="Highlights", sort_order=0),
        Section(uid="breaking", title="Breaking Changes", sort_order=1),
        Section(uid="security", title="Security Changes", sort_order=2),
        Section(uid="deprecations", title="Deprecations", sort_order=3),
        Section(uid="features", title="New Features", sort_order=4),
        Section(uid="bugfixes", title="Bug Fixes", sort_order=5),
        Section(uid="dependencies", title="Dependencies", sort_order=6),
        Section(uid="other", title="Other Changes", sort_order=7),
        Section(uid="documentation", title="Documentation", sort_order=8),
        Section(uid="internal", title="Internal Changes", sort_order=9),
    ]
)
ChangoSectionChangeNote.OWNER = "Bibo-Joshi"
ChangoSectionChangeNote.REPOSITORY = "chango"

new_chango_instance = DirectoryChanGo(
    change_note_type=ChangoSectionChangeNote,
    version_note_type=SectionVersionNote,
    version_history_type=HeaderVersionHistory,
    scanner=new_version_scanner,
)

version_scanner = BackwardCompatibleVersionScanner(
    scanners=(new_version_scanner, legacy_version_scanner)
)
chango_instance = BackwardCompatibleChanGo(
    main_instance=new_chango_instance, legacy_instances=(legacy_chango_instance,)
)



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import string
from typing import override

from ..abc import VersionHistory, VersionNote
from ..constants import MarkupLanguage
from ..error import UnsupportedMarkupError


class HeaderVersionHistory[VNT: VersionNote](VersionHistory[VNT]):
    """A simple version history implementation that renders version notes by prefixing them with
    the version UID as header, followed by the release date if available.
    """

    @override
    def render(self, markup: str) -> str:
        """Does the rendering.

        Tip:
            Version notes are automatically sorted by release date before rendering. If unreleased
            changes are present, they are rendered first.

        Important:
            Currently, only Markdown, HTML and reStructuredText are supported as markup languages.

        Args:
            markup (:obj:`str`): The markup language to use for rendering.

        Returns:
            :obj:`str`: The rendered version history.

        Raises:
            :exc:`~chango.error.UnsupportedMarkupError`: If the ``markup`` parameter does not
                coincide with :attr:`~chango.constants.MarkupLanguage.MARKDOWN`,
                :attr:`~chango.constants.MarkupLanguage.HTML`, or
                :attr:`~chango.constants.MarkupLanguage.RESTRUCTUREDTEXT`
        """
        released_notes = list(filter(lambda note: note.version, self.values()))
        changes = sorted(
            released_notes,
            key=lambda note: note.date,  # type: ignore[arg-type,return-value]
            reverse=True,
        )
        match markup:
            case MarkupLanguage.MARKDOWN:
                tpl_str = "# $uid\n*$date*\n\n$comment"
            case MarkupLanguage.HTML:
                tpl_str = "<h1>$uid</h1>\n<i>$date</i>\n\n$comment"
            case MarkupLanguage.RESTRUCTUREDTEXT:
                tpl_str = "$uid\n$rst_underline\n*$date*\n\n$comment"
            case _:
                raise UnsupportedMarkupError(
                    f"Got unsupported markup '{markup}', can only render Markdown, HTML, "
                    f"and reStructuredText"
                )

        if None in self:
            changes.insert(0, self[None])

        template = string.Template(tpl_str)
        return "\n\n".join(
            template.substitute(
                uid=note.uid or "Unreleased",  # type: ignore[truthy-function]
                rst_underline="=" * len(note.uid or "Unreleased"),  # type: ignore[truthy-function,arg-type]
                date=(
                    "unknown" if (note.date is None) else note.date.isoformat()  # type: ignore[attr-defined]
                ),
                comment=note.render(markup),
            )
            for note in changes
        )



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

import datetime as dtm

import pytest
from click import UsageError

from chango import Version
from tests.cli.conftest import ReuseCliRunner


class TestRelease:
    @pytest.mark.parametrize("has_unreleased", [True, False])
    def test_release_basic(self, cli: ReuseCliRunner, mock_chango_instance, has_unreleased):
        mock_chango_instance.release.return_value = has_unreleased
        result = cli.invoke(args=["release", "--uid", "1.0.0", "--date", "2024-01-01"])

        assert result.check_exit_code()
        assert result.stdout == (
            "Released version 1.0.0 on 2024-01-01\n"
            if has_unreleased
            else "No unreleased changes found.\n"
        )

        assert len(mock_chango_instance.release.call_args_list) == 1
        assert mock_chango_instance.release.call_args_list[0].args == (
            Version("1.0.0", dtm.date(2024, 1, 1)),
        )

    def test_release_default_date(self, cli: ReuseCliRunner, mock_chango_instance):
        result = cli.invoke(args=["release", "--uid", "1.0.0"])
        assert result.check_exit_code()
        assert result.stdout == f"Released version 1.0.0 on {dtm.date.today()}\n"

        assert len(mock_chango_instance.release.call_args_list) == 1
        assert mock_chango_instance.release.call_args_list[0].args == (
            Version("1.0.0", dtm.date.today()),
        )

    def test_release_invalid_date(self, cli: ReuseCliRunner, mock_chango_instance):
        result = cli.invoke(args=["release", "--uid", "1.0.0", "--date", "invalid"])
        assert result.check_exit_code(UsageError.exit_code)
        assert "Invalid value for '--date'" in result.stderr
        assert len(mock_chango_instance.release.call_args_list) == 0



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from contextlib import nullcontext
from pathlib import Path

import pytest
from click import UsageError

from tests.auxil.files import PROJECT_ROOT_PATH, data_path, temporary_chdir
from tests.cli.conftest import ReuseCliRunner


class TestConfig:
    @pytest.mark.parametrize(
        "path",
        [
            None,
            data_path("config/pyproject.toml"),
            data_path("config/pyproject.toml").relative_to(Path.cwd(), walk_up=True),
            data_path("config"),
        ],
        ids=["None", "absolute", "relative", "directory"],
    )
    def test_show_path_selection(self, cli: ReuseCliRunner, path):
        with temporary_chdir(data_path("config")) if path is None else nullcontext():
            args = ["config", "--path", str(path), "show"] if path else ["config", "show"]

            result = cli.invoke(args)
            assert result.check_exit_code(0)

            assert str(data_path("config/pyproject.toml")) in result.stdout
            assert 'sys_path = "/abs/sys_path"' in result.stdout
            assert (
                'chango_instance = { name= "name", module = "module", package = "package" }'
                in result.stdout
            )

    def test_show_path_not_found(self, cli: ReuseCliRunner):
        path = Path("nonexistent").absolute()
        result = cli.invoke(["config", "--path", str(path), "show"])
        assert result.check_exit_code(UsageError.exit_code)
        assert f"File not found: {path!s}" in result.stderr

    def test_show_invalid_toml(self, cli: ReuseCliRunner, tmp_path):
        path = tmp_path / "pyproject.toml"
        path.write_text("invalid toml")

        with temporary_chdir(tmp_path):
            result = cli.invoke(["config", "--path", str(path), "show"])
            assert result.check_exit_code(UsageError.exit_code)
            assert "Failed to parse the configuration file" in result.stderr

    def test_show_no_chango_config(self, cli: ReuseCliRunner, tmp_path):
        path = tmp_path / "pyproject.toml"
        path.write_text("[tool.other]")

        with temporary_chdir(tmp_path):
            result = cli.invoke(["config", "--path", str(path), "show"])
            assert result.check_exit_code(UsageError.exit_code)
            assert "No configuration found for chango" in result.stderr

    def test_validate_invalid_chango_config(self, cli: ReuseCliRunner, tmp_path):
        path = tmp_path / "pyproject.toml"
        path.write_text("[tool.chango]\nsys_path = 42")

        with temporary_chdir(tmp_path):
            result = cli.invoke(["config", "--path", str(path), "validate"])
            assert result.check_exit_code(UsageError.exit_code)
            assert f"Validation of config file at {path!s} failed:" in result.stderr

    def test_validate_import_error(self, cli: ReuseCliRunner, tmp_path):
        path = tmp_path / "pyproject.toml"
        path.write_text(
            "[tool.chango]\nsys_path = 'sys_path'\nchango_instance = { name= 'name', module = "
            "'module', package = 'package' }"
        )

        with temporary_chdir(tmp_path):
            result = cli.invoke(["config", "--path", str(path), "validate"])
            assert result.check_exit_code(UsageError.exit_code)
            assert "importing the ChanGo instance failed" in result.stderr

    def test_validate_success(self, cli: ReuseCliRunner):
        with temporary_chdir(PROJECT_ROOT_PATH):
            result = cli.invoke(["config", "validate"])
            assert result.check_exit_code(0)
            assert "The configuration in" in result.stdout
            assert str(PROJECT_ROOT_PATH / "pyproject.toml") in result.stdout
            assert "valid" in result.stdout



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import datetime as dtm

import pytest

from chango import Version


class TestVersion:
    def test_init(self):
        v = Version("1.2.3", dtm.date(2024, 1, 1))
        assert v.uid == "1.2.3"
        assert v.date == dtm.date(2024, 1, 1)

    def test_frozen(self):
        v = Version("1.2.3", dtm.date(2024, 1, 1))
        with pytest.raises(AttributeError):
            v.uid = "1.2.4"
        with pytest.raises(AttributeError):
            v.date = dtm.date(2024, 1, 2)



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import pytest

from chango.concrete import CommentChangeNote
from chango.constants import MarkupLanguage


class TestCommentChangeNote:
    change_note = CommentChangeNote(slug="slug", comment="comment 𝛙𝌢𑁍")

    def test_init(self):
        assert self.change_note.comment == "comment 𝛙𝌢𑁍"

    def test_file_extension(self):
        assert self.change_note.file_extension == CommentChangeNote.MARKUP

    def test_from_string(self):
        change_note = CommentChangeNote.from_string("slug", "uid", "comment")
        assert change_note.comment == "comment"

    @pytest.mark.parametrize("encoding", ["utf-8", "utf-16"])
    def test_to_bytes(self, encoding):
        assert self.change_note.to_bytes(encoding) == (
            b"comment \xf0\x9d\x9b\x99\xf0\x9d\x8c\xa2\xf0\x91\x81\x8d"
            if encoding == "utf-8"
            else (
                b"\xff\xfec\x00o\x00m\x00m\x00e\x00n\x00t\x00 "
                b'\x005\xd8\xd9\xde4\xd8"\xdf\x04\xd8M\xdc'
            )
        )

    @pytest.mark.parametrize("encoding", ["utf-8", "utf-16"])
    def test_to_string(self, encoding):
        assert self.change_note.to_string(encoding=encoding) == "comment 𝛙𝌢𑁍"

    def test_build_template(self):
        change_note = CommentChangeNote.build_template("slug", "uid")
        assert change_note.comment == "example comment"
        assert change_note.slug == "slug"
        assert change_note.uid == "uid"

    def test_build_template_no_uid(self):
        change_note = CommentChangeNote.build_template("slug")
        assert change_note.comment == "example comment"
        assert change_note.slug == "slug"
        assert isinstance(change_note.uid, str)
        assert len(change_note.uid) > 0

    def test_build_from_github_event_missing_data(self):
        with pytest.raises(ValueError, match="required data"):
            CommentChangeNote.build_from_github_event({})

    def test_build_from_github_event_unsupported_language(self, monkeypatch):
        monkeypatch.setattr(CommentChangeNote, "MARKUP", "unsupported markup language")
        with pytest.raises(ValueError, match="unsupported markup language"):
            CommentChangeNote.build_from_github_event(
                {
                    "pull_request": {
                        "html_url": "https://example.com/pull/42",
                        "number": 42,
                        "title": "example title",
                    }
                }
            )

    @pytest.mark.parametrize(
        ("language", "expected"),
        [
            (MarkupLanguage.TEXT, "example title (https://example.com/pull/42)"),
            (MarkupLanguage.MARKDOWN, "example title ([#42](https://example.com/pull/42))"),
            (
                MarkupLanguage.RESTRUCTUREDTEXT,
                "example title (`#42 <https://example.com/pull/42>`_)",
            ),
            (MarkupLanguage.HTML, 'example title (<a href="https://example.com/pull/42">#42</a>)'),
        ],
    )
    def test_build_from_github_event(self, language, expected, monkeypatch):
        monkeypatch.setattr(CommentChangeNote, "MARKUP", language)
        event_data = {
            "pull_request": {
                "html_url": "https://example.com/pull/42",
                "number": 42,
                "title": "example title",
            }
        }

        change_note = CommentChangeNote.build_from_github_event(event_data)
        assert change_note.comment == expected
        assert change_note.slug == "0042"



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import abc
import datetime as dtm
import warnings
from collections.abc import Iterator, MutableMapping
from typing import TYPE_CHECKING, overload

from .._utils.filename import FileName
from ..abc._changenote import ChangeNote
from ..error import ValidationError

if TYPE_CHECKING:
    from chango import Version


class VersionNote[CNT: ChangeNote, V: (Version, None)](MutableMapping[str, CNT], abc.ABC):
    """Abstract base class for a version note describing the set of changes in a software project
    for a single version.

    Hint:
        Objects of this class can be used as :class:`~collections.abc.MutableMapping`, where the
        keys are the unique identifiers (or file names) of the change notes and the values are the
        change notes themselves.

    Warning:
        To ensure that the changes in this version are displayed in the correct order, the change
        notes should be added in the order they were made. Manual reordering of the change notes
        may interfere with the order in which they are displayed.

    Args:
        version (:class:`~chango.Version` | :obj:`None`): The version of the software project this
            note is for or May be :obj:`None` if the version is not yet released.

    Attributes:
        version: (:class:`~chango.Version` | :obj:`None`): The version of the software project this
            note is for or May be :obj:`None` if the version is not yet released.
    """

    def __init__(self, version: V) -> None:
        self.version: V = version
        self._change_notes: dict[str, CNT] = {}

    def __delitem__(self, key: str, /) -> None:
        try:
            del self._change_notes[key]
        except KeyError:
            try:
                del self._change_notes[FileName.from_string(key).uid]
            except ValidationError:
                raise KeyError(key) from None

    def __getitem__(self, key: str, /) -> CNT:
        try:
            return self._change_notes[key]
        except KeyError:
            try:
                return self._change_notes[FileName.from_string(key).uid]
            except ValidationError:
                raise KeyError(key) from None

    def __iter__(self) -> Iterator[str]:
        return iter(self._change_notes)

    def __len__(self) -> int:
        return len(self._change_notes)

    def __setitem__(self, key: str, value: CNT, /) -> None:
        if key != value.uid:
            warnings.warn(
                f"Key {key!r} does not match change note UID {value.uid!r}. Using the UID as key.",
                stacklevel=2,
            )
        self._change_notes[value.uid] = value

    @overload
    def uid(self: "VersionNote[CNT, Version]") -> str: ...

    @overload
    def uid(self: "VersionNote[CNT, None]") -> None: ...

    @property
    def uid(self) -> str | None:
        """Convenience property for the version UID.

        Returns:
            :obj:`str` | :obj:`None`: The UID of :attr:`version` if available, :obj:`None`
                otherwise.
        """
        if self.version is None:
            return None
        return self.version.uid

    @overload
    def date(self: "VersionNote[CNT, Version]") -> dtm.date: ...

    @overload
    def date(self: "VersionNote[CNT, None]") -> None: ...

    @property
    def date(self) -> dtm.date | None:
        """Convenience property for the version UID.

        Returns:
            :class:`datetime.date` | :obj:`None`: The release date of :attr:`version` if available,
                :obj:`None` otherwise.
        """
        if self.version is None:
            return None
        return self.version.date

    def add_change_note(self, change_note: CNT) -> None:
        """Add a change note to the version note.

        Args:
            change_note (:class:`CNT <typing.TypeVar>`): The :class:`~chango.abc.ChangeNote` note
                to add.
        """
        self[change_note.uid] = change_note

    def remove_change_note(self, change_note: CNT) -> None:
        """Remove a change note from the version note.

        Args:
            change_note (:class:`CNT <typing.TypeVar>`): The :class:`~chango.abc.ChangeNote` note
                to remove.
        """
        del self[change_note.uid]

    @abc.abstractmethod
    def render(self, markup: str) -> str:
        """Render the version note as a string.

        Args:
            markup (:obj:`str`): The markup language to use for rendering. If the markup language
                is not supported, an :exc:`~chango.error.UnsupportedMarkupError` should be raised.

        Returns:
            :obj:`str`: The rendered version note.
        """



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
"""This module provides the functionality to load the configuration for the ChanGo CLI."""

__all__ = ["ChanGoConfig", "ChanGoInstanceConfig", "get_chango_instance"]

import functools
from typing import Any

from chango.abc import ChanGo

from .._utils.types import PathLike
from ._models import ChanGoConfig, ChanGoInstanceConfig


@functools.lru_cache(maxsize=256)
def get_chango_instance(path: PathLike | None = None) -> ChanGo[Any, Any, Any, Any]:
    """Get the :class:`~chango.abc.ChanGo` instance specified in the configuration file.
    Uses LRU caching to avoid reloading the configuration file multiple times.

    Args:
        path (:class:`~pathlib.Path` | :obj:`str` | :obj:`None`, optional): The path to the
            configuration file as passed to :meth:`ChanGoConfig.load`.

    Returns:
        :class:`~chango.abc.ChanGo`: The instance of the :class:`~chango.abc.ChanGo` class
            specified in the configuration file.
    """
    return ChanGoConfig.load(path).import_chango_instance()



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

__all__ = ["app"]

from typing import Annotated

import typer

from chango.config import get_chango_instance

from ..constants import MarkupLanguage
from .utils.types import MARKUP, OUTPUT_FILE

app = typer.Typer(help="Generate reports for one or multiple versions.")


@app.command()
def version(
    uid: Annotated[
        str,
        typer.Option(
            help=(
                "The unique identifier of the version to report on. Leave empty for unreleased "
                "changes."
            ),
            show_default=False,
        ),
    ],
    markup: MARKUP = MarkupLanguage.MARKDOWN,
    output: OUTPUT_FILE = None,
) -> None:
    """Print a report of the change notes for a specific version."""
    version_note = get_chango_instance().load_version_note(uid)
    text = version_note.render(markup=markup)
    if output:
        output.write_text(text)
        typer.echo(f"Report written to {output}")
    else:
        typer.echo(text)


@app.command()
def history(markup: MARKUP = MarkupLanguage.MARKDOWN, output: OUTPUT_FILE = None) -> None:
    """Print a report of the version history."""
    version_history = get_chango_instance().load_version_history()
    text = version_history.render(markup=markup)
    if output:
        output.write_text(text)
        typer.echo(f"Report written to {output}")
    else:
        typer.echo(text)



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import logging
from collections.abc import Callable, Sequence
from pathlib import Path
from string import Template
from typing import Annotated

import pytest
import shortuuid
from _pytest.tmpdir import TempPathFactory
from sphinx.testing.util import SphinxTestApp
from sphinx.util.logging import pending_warnings

from chango import __version__
from chango._utils.types import PathLike
from chango.constants import MarkupLanguage
from tests.auxil.files import data_path, path_to_python_string
from tests.sphinx_ext.conftest import MockChanGo, MockStorage

MAKE_APP_TYPE = Callable[..., SphinxTestApp]


class SphinxBuildError(RuntimeError):
    pass


class TestSphinxExt:
    SPHINX_EXT_TEST_ROOT = data_path("sphinx_ext")

    @staticmethod
    def assert_successful_build(app: SphinxTestApp):
        with pending_warnings() as mem_handler:
            app.build()

            for record in mem_handler.buffer:
                if record.levelno == logging.ERROR:
                    raise SphinxBuildError(record.getMessage())

    @classmethod
    def create_template(
        cls,
        *,
        tmp_path_factory: TempPathFactory,
        conf_value_insert: str | None = None,
        directive_insert: str = ".. chango::",
    ) -> Path:
        uid = shortuuid.uuid()
        src_dir = f"test-{uid}"
        tmp_dir = tmp_path_factory.mktemp(src_dir)

        conf_template = Template(
            (cls.SPHINX_EXT_TEST_ROOT / "conf_value.py.template").read_text(encoding="utf-8")
        )
        index_template = Template(
            (cls.SPHINX_EXT_TEST_ROOT / "index.rst.template").read_text(encoding="utf-8")
        )

        (tmp_dir / "conf.py").write_text(
            conf_template.substitute(chango_pyproject_toml_path=conf_value_insert),
            encoding="utf-8",
        )
        (tmp_dir / "index.rst").write_text(
            index_template.substitute(directive=directive_insert), encoding="utf-8"
        )

        return tmp_dir

    @staticmethod
    def compute_chango_pyproject_toml_path_insert(
        path: str | Path, path_representation: PathLike
    ) -> str:
        if path == "explicit_none":
            return "chango_pyproject_toml_path = None"
        if path is None:
            return ""
        return f"chango_pyproject_toml_path = {path_to_python_string(path, path_representation)}"

    @pytest.mark.parametrize(
        "path",
        [
            None,
            "explicit_none",
            data_path("config/pyproject.toml"),
            data_path("config/pyproject.toml").relative_to(Path.cwd(), walk_up=True),
            data_path("config"),
        ],
        ids=["None", "explicit_none", "absolute", "relative", "directory"],
    )
    @pytest.mark.parametrize("path_representation", [str, Path])
    def test_chango_pyproject_toml_path_valid(
        self, path, path_representation, make_app: MAKE_APP_TYPE, tmp_path_factory: TempPathFactory
    ):
        app = make_app(
            srcdir=self.create_template(
                conf_value_insert=self.compute_chango_pyproject_toml_path_insert(
                    path, path_representation
                ),
                tmp_path_factory=tmp_path_factory,
            )
        )

        if path in ("explicit_none", None):
            assert app.config.chango_pyproject_toml_path is None
        else:
            assert (
                app.config.chango_pyproject_toml_path == path.as_posix()
                if path_representation is str
                else path
            )

    @pytest.mark.parametrize("path", [1, {"key": "value"}, [1, 2, 3]], ids=["int", "dict", "list"])
    def test_chango_pyproject_toml_path_invalid(
        self, path, make_app: MAKE_APP_TYPE, tmp_path_factory: TempPathFactory
    ):
        insert = f"chango_pyproject_toml_path = {path!r}"

        with pytest.raises(
            TypeError, match="Expected 'chango_pyproject_toml_path' to be a string or Path"
        ):
            make_app(
                srcdir=self.create_template(
                    conf_value_insert=insert, tmp_path_factory=tmp_path_factory
                )
            )

    def test_metadata(self, app: SphinxTestApp):
        assert app.extensions["chango.sphinx_ext"].version == __version__
        assert app.extensions["chango.sphinx_ext"].parallel_read_safe is True
        assert app.extensions["chango.sphinx_ext"].parallel_write_safe is True

    @pytest.mark.parametrize(
        "path",
        [
            None,
            "explicit_none",
            data_path("config/pyproject.toml"),
            data_path("config/pyproject.toml").relative_to(Path.cwd(), walk_up=True),
            data_path("config"),
        ],
        ids=["None", "explicit_none", "absolute", "relative", "directory"],
    )
    @pytest.mark.parametrize("path_representation", [str, Path])
    def test_directive_chango_instance_loading(
        self,
        make_app: MAKE_APP_TYPE,
        path,
        path_representation,
        tmp_path_factory: TempPathFactory,
        cg_config_mock,
    ):
        app = make_app(
            srcdir=self.create_template(
                conf_value_insert=self.compute_chango_pyproject_toml_path_insert(
                    path, path_representation
                ),
                tmp_path_factory=tmp_path_factory,
            )
        )
        self.assert_successful_build(app)
        received_sys_path = cg_config_mock.get().sys_path

        if path in ("explicit_none", None):
            assert received_sys_path is None
        else:
            assert received_sys_path == path

    @pytest.mark.parametrize(
        "headline", [None, "This is a headline"], ids=["no_headline", "headline"]
    )
    def test_directive_rendering_basic(
        self, cg_config_mock, make_app: MAKE_APP_TYPE, tmp_path_factory: TempPathFactory, headline
    ):
        directive = f".. chango:: {headline}" if headline else ".. chango::"

        app = make_app(
            srcdir=self.create_template(
                directive_insert=directive, tmp_path_factory=tmp_path_factory
            )
        )

        self.assert_successful_build(app)

        index = app.outdir.joinpath("index.html")
        assert index.exists()

        content = index.read_text(encoding="utf-8")
        assert cg_config_mock.rendered_content in content

        if headline:
            assert f"<h1>{headline}" in content
        else:
            assert f"<h1>{headline}" not in content

    def test_directive_rendering_passed_markup_language(self, cg_config_mock, app: SphinxTestApp):
        self.assert_successful_build(app)
        received_args = cg_config_mock.get().chango.version_history.received_args
        received_kwargs = cg_config_mock.get().chango.version_history.received_kwargs
        assert received_args == (MarkupLanguage.RESTRUCTUREDTEXT,)
        assert received_kwargs == {}

    def test_argument_passing_basic(
        self,
        cg_config_mock: MockStorage,
        make_app: MAKE_APP_TYPE,
        tmp_path_factory: TempPathFactory,
    ):
        directive = """
.. chango::
   :start_from: "start_from"
   :end_at: "end_at"
"""
        app = make_app(
            srcdir=self.create_template(
                directive_insert=directive, tmp_path_factory=tmp_path_factory
            )
        )

        self.assert_successful_build(app)
        received_kwargs = cg_config_mock.get().chango.received_kwargs
        assert received_kwargs == {"start_from": "start_from", "end_at": "end_at"}

    def test_argument_passing_unknown_option(
        self, make_app: MAKE_APP_TYPE, tmp_path_factory: TempPathFactory
    ):
        directive = """
.. chango::
   :unknown:
"""
        app = make_app(
            srcdir=self.create_template(
                directive_insert=directive, tmp_path_factory=tmp_path_factory
            )
        )

        with pytest.raises(SphinxBuildError, match='unknown option: "unknown"'):
            self.assert_successful_build(app)

    def test_argument_passing_json_data(
        self,
        cg_config_mock: MockStorage,
        make_app: MAKE_APP_TYPE,
        tmp_path_factory: TempPathFactory,
    ):
        directive = """
.. chango::
   :start_from: {"key": "value"}
   :end_at: [false, true, null]
"""
        app = make_app(
            srcdir=self.create_template(
                directive_insert=directive, tmp_path_factory=tmp_path_factory
            )
        )

        self.assert_successful_build(app)
        received_kwargs = cg_config_mock.get().chango.received_kwargs
        assert received_kwargs == {"start_from": {"key": "value"}, "end_at": [False, True, None]}

    def test_argument_passing_invalid_json_data(
        self, make_app: MAKE_APP_TYPE, tmp_path_factory: TempPathFactory
    ):
        directive = """
.. chango::
    :start_from: {"key": "value}
    """
        app = make_app(
            srcdir=self.create_template(
                directive_insert=directive, tmp_path_factory=tmp_path_factory
            )
        )

        with pytest.raises(SphinxBuildError, match="must be a JSON-loadable value"):
            self.assert_successful_build(app)

    def test_argument_passing_missing_value(
        self, make_app: MAKE_APP_TYPE, tmp_path_factory: TempPathFactory
    ):
        directive = """
.. chango::
    :start_from:
    """
        app = make_app(
            srcdir=self.create_template(
                directive_insert=directive, tmp_path_factory=tmp_path_factory
            )
        )

        with pytest.raises(SphinxBuildError, match="must be a JSON-loadable value"):
            self.assert_successful_build(app)

    def test_argument_passing_custom_signature(
        self,
        cg_config_mock: MockStorage,
        make_app: MAKE_APP_TYPE,
        tmp_path_factory: TempPathFactory,
        monkeypatch,
    ):
        validator_kwargs = {}

        def sequence_validator(value: str | None) -> Sequence[int]:
            validator_kwargs["sequence_validator"] = value
            return tuple(map(int, value.split(",")))

        def flag_validator(value: str | None) -> bool:
            validator_kwargs["flag_validator"] = value
            return True

        original_load_version_history = MockChanGo.load_version_history

        def load_version_history(
            *args,
            start_from: str | None = None,
            end_at: str | None = None,
            json_dict_arg: dict[str, str] | None = None,
            sequence_arg: Annotated[Sequence[int], sequence_validator] = (1, 2, 3),
            flag_arg: Annotated[bool, flag_validator] = False,
        ):
            return original_load_version_history(
                *args,
                start_from=start_from,
                end_at=end_at,
                json_dict_arg=json_dict_arg,
                sequence_arg=sequence_arg,
                flag_arg=flag_arg,
            )

        monkeypatch.setattr(MockChanGo, "load_version_history", load_version_history)

        directive = """
.. chango::
    :json_dict_arg: {"key": "value"}
    :sequence_arg: 1,2,3
    :flag_arg:
    """

        app = make_app(
            srcdir=self.create_template(
                directive_insert=directive, tmp_path_factory=tmp_path_factory
            )
        )

        self.assert_successful_build(app)

        received_kwargs = cg_config_mock.get().chango.received_kwargs
        assert received_kwargs == {
            "json_dict_arg": {"key": "value"},
            "sequence_arg": (1, 2, 3),
            "flag_arg": True,
            "start_from": None,
            "end_at": None,
        }
        assert validator_kwargs == {"sequence_validator": "1,2,3", "flag_validator": None}



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import contextlib
from collections.abc import Collection
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, override

from .._utils.types import VUIDInput
from ..abc import ChangeNote, ChanGo, VersionHistory, VersionNote
from ..action import ChanGoActionData
from ..error import ChanGoError
from ._backwardcompatibleversionscanner import BackwardCompatibleVersionScanner

if TYPE_CHECKING:
    from .. import Version


class BackwardCompatibleChanGo[VHT: VersionHistory, VNT: VersionNote, CNT: ChangeNote](
    ChanGo["BackwardCompatibleVersionScanner", VHT, VNT, CNT]
):
    """An Implementation of the :class:`~chango.abc.ChanGo` interface that wraps multiple
    other implementations of :class:`~chango.abc.ChanGo`.
    The purpose of this class is to ease transition between different version note formats in
    a project.

    Args:
        main_instance(:class:`~chango.abc.ChanGo`): The :class:`~chango.abc.ChanGo` instance that
            should be used for new version notes.
        legacy_instances(Collection[:class:`~chango.abc.ChanGo`]): A collection of
            :class:`~chango.abc.ChanGo` instances that should be used for loading older version
            notes.
    """

    def __init__(
        self,
        main_instance: "ChanGo[Any,VHT, VNT, CNT]",
        legacy_instances: Collection[ChanGo[Any, Any, Any, Any]],
    ):
        self._main_instance = main_instance
        self._legacy_instances = tuple(legacy_instances)
        self._scanner = BackwardCompatibleVersionScanner(
            (main_instance.scanner, *(chango.scanner for chango in self._legacy_instances))
        )

    @property
    @override
    def scanner(self) -> "BackwardCompatibleVersionScanner":
        """The :class:`~chango.concrete.BackwardCompatibleVersionScanner` instance that is used
        by this :class:`BackwardCompatibleChanGo`.

        Hint:
            The scanner is a composite of the scanners of
            :paramref:`~BackwardCompatibleChanGo.main_instance` and
            :paramref:`~BackwardCompatibleChanGo.legacy_instance`.
        """
        return self._scanner

    @override
    def build_template_change_note(self, slug: str, uid: str | None = None) -> CNT:
        """Calls :meth:`~chango.abc.ChanGo.build_template_change_note` on
        :paramref:`~BackwardCompatibleChanGo.main_instance`.
        """
        return self._main_instance.build_template_change_note(slug, uid)

    @override
    def build_version_note(self, version: Optional["Version"]) -> VNT:
        """Calls :meth:`~chango.abc.ChanGo.build_version_note`
        on :paramref:`~BackwardCompatibleChanGo.main_instance` or one of the legacy
        instances depending on the result of :meth:`~chango.abc.VersionScanner.is_available`.
        """
        for chango in (self._main_instance, *self._legacy_instances):
            if chango.scanner.is_available(version):
                return chango.build_version_note(version)
        raise ChanGoError(f"Version {version} not found")

    @override
    def build_version_history(self) -> VHT:
        """Calls :meth:`~chango.abc.ChanGo.build_version_history`
        on :paramref:`~BackwardCompatibleChanGo.main_instance`.
        """
        return self._main_instance.build_version_history()

    @override
    def load_change_note(self, uid: str) -> CNT:
        """Load a change note with the given identifier.
        Tries to load the change note from the main chango first and then from the legacy changos.
        """
        for chango in (self._main_instance, *self._legacy_instances):
            with contextlib.suppress(ChanGoError):
                return chango.load_change_note(uid)
        raise ChanGoError(f"Change note with uid {uid} not found")

    @override
    def get_write_directory(self, change_note: CNT | str, version: VUIDInput) -> Path:
        """Calls :meth:`~chango.abc.ChanGo.get_write_directory`
        on :paramref:`~BackwardCompatibleChanGo.main_instance`.
        """
        return self._main_instance.get_write_directory(change_note, version)

    def build_github_event_change_note(
        self, event: dict[str, Any], data: dict[str, Any] | ChanGoActionData | None = None
    ) -> CNT | None:
        """Calls :meth:`~chango.abc.ChanGo.build_github_event_change_note`
        on :paramref:`~BackwardCompatibleChanGo.main_instance`.
        """
        return self._main_instance.build_github_event_change_note(event, data)



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import datetime as dtm
import functools
import shutil
import subprocess
from pathlib import Path

import pytest

import chango as chango_module
from chango import Version
from chango.abc import ChanGo
from chango.concrete import (
    CommentChangeNote,
    CommentVersionNote,
    DirectoryChanGo,
    DirectoryVersionScanner,
    HeaderVersionHistory,
)
from chango.error import ChanGoError
from chango.helpers import ensure_uid
from tests.auxil.files import data_path


@pytest.fixture
def scanner() -> DirectoryVersionScanner:
    return DirectoryVersionScanner(TestChanGo.DATA_ROOT, "unreleased")


@pytest.fixture
def chango(scanner) -> DirectoryChanGo:
    return DirectoryChanGo(
        change_note_type=CommentChangeNote,
        version_note_type=CommentVersionNote,
        version_history_type=HeaderVersionHistory,
        scanner=scanner,
    )


@pytest.fixture
def chango_no_unreleased() -> DirectoryChanGo:
    return DirectoryChanGo(
        change_note_type=CommentChangeNote,
        version_note_type=CommentVersionNote,
        version_history_type=HeaderVersionHistory,
        scanner=DirectoryVersionScanner(TestChanGo.DATA_ROOT, "no-unreleased"),
    )


@pytest.fixture
def cache_invalidation_tracker():
    class Tracker:
        def __init__(self):
            self.invalidate_caches = False
            self.super_call = None

        def __call__(self, *args, **kwargs):
            self.invalidate_caches = True
            if self.super_call:
                self.super_call(*args, **kwargs)

        @property
        def was_called(self):
            return self.invalidate_caches

        def set_super(self, super_call):
            self.super_call = super_call
            return self

    return Tracker()


class TestChanGo:
    """Since Chango is an abstract base class, we are testing with DirectoryChanGo as a simple
    implementation.

    Note that we do *not* test abstract methods, as that is the responsibility of the concrete
    implementations.
    """

    DATA_ROOT = data_path("directoryversionscanner")

    @pytest.mark.parametrize(
        "version",
        [
            None,
            "1.1",
            Version("1.1", dtm.date(2024, 1, 1)),
            "1.2",
            Version("1.2", dtm.date(2024, 1, 2)),
            Version("new-version", dtm.date(2024, 1, 17)),
        ],
    )
    @pytest.mark.parametrize("encoding", ["utf-8", "utf-16"])
    @pytest.mark.parametrize(
        "has_git", [pytest.param(True, id="with-git"), pytest.param(False, id="without-git")]
    )
    def test_write_change_note(
        self, chango, version, monkeypatch, encoding, cache_invalidation_tracker, has_git
    ):
        # Unfortunately, testing the git-available part is not easily possible without using
        # some of the internal utils and also not with directly running git. This is because
        # a) the availability of git is cached and there is no public interface to reset it
        # b) mocking subprocess before the module is imported is not easily possible
        # c) actually running `git add` is hard to reset
        # Since `chango._utils.files` is not part of the public API, we settle for testing
        # with the private interfaces.
        chango_module._utils.files._GIT_HELPER.git_available = None

        def check_call(args, *_, **__):
            assert args[:2] == ["git", "add"]
            if not has_git:
                raise subprocess.CalledProcessError(1, "git add")

        monkeypatch.setattr("chango._utils.files.subprocess.check_call", check_call)

        if version is None:
            expected_path = chango.scanner.unreleased_directory
        else:
            version_uid = ensure_uid(version)
            if version_uid == "new-version":
                expected_path = chango.scanner.base_directory / "new-version_2024-01-17"
            else:
                day = int(version_uid.split(".")[-1])
                expected_path = chango.scanner.base_directory / f"{version_uid}_2024-01-0{day}"

        existed = expected_path.is_dir()

        def to_file(*_, **kwargs):
            assert kwargs.get("encoding") == encoding
            assert kwargs.get("directory") == expected_path

        note = chango.build_template_change_note("this-is-a-new-slug")
        monkeypatch.setattr(note, "to_file", to_file)
        monkeypatch.setattr(chango.scanner, "invalidate_caches", cache_invalidation_tracker)

        for _ in range(3):
            # run multiple times to cover all paths in _GIT_HELPER
            try:
                chango.write_change_note(note, version, encoding=encoding)
                assert cache_invalidation_tracker.was_called
            finally:
                if not existed and expected_path.is_dir():
                    shutil.rmtree(expected_path)

    def test_write_change_note_new_string_version(self, chango):
        note = chango.build_template_change_note("this-is-a-new-slug")
        with pytest.raises(ChanGoError, match="'new-version-uid' not available"):
            chango.write_change_note(note, "new-version-uid")

    def test_load_version_note_unavailable(self, chango):
        with pytest.raises(ChanGoError, match=r"Version '1.4' not available."):
            chango.load_version_note("1.4")

    @pytest.mark.parametrize(
        "version",
        [
            None,
            "1.1",
            Version("1.1", dtm.date(2024, 1, 1)),
            "1.2",
            Version("1.2", dtm.date(2024, 1, 2)),
        ],
    )
    def test_load_version_note(self, chango, version):
        version_note = chango.load_version_note(version)

        version_uid = ensure_uid(version)
        expected_uids = {
            f"uid_{(version_uid or 'ur').replace('.', '-')}_{idx}" for idx in range(3)
        }

        assert version_note.uid == version_uid
        assert version_note.date == (
            dtm.date(2024, 1, int(version_uid.split(".")[-1])) if version else None
        )
        assert set(version_note) == expected_uids

    @pytest.mark.parametrize(
        ("start_from", "end_at"),
        [(None, None), (None, "1.2"), ("1.3", None), ("1.2", "1.3"), ("1.3", "1.3")],
    )
    def test_load_version_history(self, chango, start_from, end_at):
        lower_idx = int(start_from.split(".")[-1]) if start_from else 1
        upper_idx = int(end_at.split(".")[-1]) + 1 if end_at else 4

        versions = {
            Version(f"1.{idx}", dtm.date(2024, 1, idx)) for idx in range(lower_idx, upper_idx)
        }
        if not end_at:
            versions |= {Version("1.3.1", dtm.date(2024, 1, 3)), None}
        version_history = chango.load_version_history(start_from, end_at)

        assert set(version_history) == set(map(ensure_uid, versions))
        for version in versions:
            assert version_history[ensure_uid(version)].date == (version.date if version else None)
            assert version_history[ensure_uid(version)].version == version

    def test_release_no_unreleased_changes(
        self, chango_no_unreleased: ChanGo, monkeypatch, cache_invalidation_tracker
    ):
        monkeypatch.setattr(
            chango_no_unreleased.scanner, "invalidate_caches", cache_invalidation_tracker
        )

        version = Version("1.4", dtm.date(2024, 1, 4))
        assert not chango_no_unreleased.release(version)
        assert not chango_no_unreleased.scanner.is_available(version)
        assert not cache_invalidation_tracker.was_called

    @pytest.mark.parametrize(
        "has_git", [pytest.param(True, id="with-git"), pytest.param(False, id="without-git")]
    )
    def test_release(self, chango, cache_invalidation_tracker, monkeypatch, has_git):
        # Unfortunately, testing the git-available part is not easily possible without using
        # some of the internal utils and also not with directly running git. This is because
        # a) the availability of git is cached and there is no public interface to reset it
        # b) mocking subprocess before the module is imported is not easily possible
        # c) actually running `git mv` is harder to reset than just using the pathlib move
        # Since `chango._utils.files` is not part of the public API, we settle for testing
        # with the private interfaces.
        chango_module._utils.files._GIT_HELPER.git_available = None

        def check_call(args, *_, **__):
            assert args[:2] == ["git", "mv"]
            if not has_git:
                raise subprocess.CalledProcessError(1, "git mv")

            source, destination = args[2], args[3]
            Path(source).rename(destination)

        monkeypatch.setattr("chango._utils.files.subprocess.check_call", check_call)

        version = Version("1.4", dtm.date(2024, 1, 4))
        expected_path = chango.scanner.base_directory / "1.4_2024-01-04"
        expected_files = {
            (file.name, file.read_bytes())
            for file in (self.DATA_ROOT / "unreleased").iterdir()
            if file.name != "not-a-change-note.txt"
        }

        monkeypatch.setattr(
            chango.scanner,
            "invalidate_caches",
            cache_invalidation_tracker.set_super(chango.scanner.invalidate_caches),
        )

        try:
            assert chango.release(version)
            assert cache_invalidation_tracker.was_called
            assert chango.scanner.is_available(version)
            assert chango.scanner.get_version(version.uid) == version

            assert expected_path.is_dir()
            assert {
                (file.name, file.read_bytes()) for file in expected_path.iterdir()
            } == expected_files
        finally:
            for file_name, file_content in expected_files:
                (self.DATA_ROOT / "unreleased" / file_name).write_bytes(file_content)
            if expected_path.is_dir():
                shutil.rmtree(expected_path)

    def test_release_same_directory(self, chango, monkeypatch, cache_invalidation_tracker):
        def get_write_directory(*_, **__):
            return chango.scanner.unreleased_directory

        monkeypatch.setattr(chango, "get_write_directory", get_write_directory)
        monkeypatch.setattr(
            chango.scanner,
            "invalidate_caches",
            cache_invalidation_tracker.set_super(chango.scanner.invalidate_caches),
        )

        version = Version("1.4", dtm.date(2024, 1, 4))
        expected_files = {
            (file.name, file.read_bytes())
            for file in (self.DATA_ROOT / "unreleased").iterdir()
            if file.name != "not-a-change-note.txt"
        }
        try:
            assert chango.release(version)
            assert cache_invalidation_tracker.was_called
            for file_name, file_content in expected_files:
                assert (self.DATA_ROOT / "unreleased" / file_name).read_bytes() == file_content
        except Exception:
            for file_name, file_content in expected_files:
                (self.DATA_ROOT / "unreleased" / file_name).write_bytes(file_content)

    def test_build_github_event_change_note(self, chango, monkeypatch):
        monkeypatch.setattr(
            chango,
            "build_github_event_change_note",
            functools.partial(ChanGo.build_github_event_change_note, chango),
        )
        with pytest.raises(NotImplementedError):
            chango.build_github_event_change_note({})



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import pytest

from chango.concrete.sections import (
    GitHubSectionChangeNote,
    PullRequest,
    Section,
    SectionChangeNote,
    SectionVersionNote,
)
from chango.error import UnsupportedMarkupError


class DummySectionChangeNote(
    GitHubSectionChangeNote.with_sections(
        [
            Section(uid="req_section", title="Required Section", is_required=True, sort_order=10),
            Section(uid="opt_section", title="Optional Section", render_pr_details=False),
        ]
    )
):
    OWNER = "my-username"
    REPOSITORY = "my-repo"


OtherSectionChangeNote = GitHubSectionChangeNote.with_sections(
    [Section(uid="req_section", title="Required Section", is_required=True)]
)


@pytest.fixture
def section_version_note():
    return SectionVersionNote(None, DummySectionChangeNote)


class TestSectionVersionNote:
    def test_init_invalid_markup(self):
        class InvalidSectionChangeNote(SectionChangeNote):
            MARKUP = "invalid"

        with pytest.raises(UnsupportedMarkupError, match="only supports reStructuredText"):
            SectionVersionNote(None, InvalidSectionChangeNote)

    def test_setitem_invalid_type(self, section_version_note):
        other_section_change_note = OtherSectionChangeNote(slug="slug", req_section="req_section")
        with pytest.raises(TypeError, match="Expected a"):
            section_version_note["key"] = other_section_change_note
        with pytest.raises(TypeError, match="Expected a"):
            section_version_note.add_change_note(other_section_change_note)

    def test_render_unknown_markup(self, section_version_note):
        with pytest.raises(UnsupportedMarkupError, match="unknown"):
            section_version_note.render("unknown")

    def test_render_unsupported_markup(self, section_version_note):
        with pytest.raises(UnsupportedMarkupError, match="markdown"):
            section_version_note.render("markdown")

    def test_render(self, section_version_note):
        section_version_note.add_change_note(
            DummySectionChangeNote(
                slug="slug1",
                uid="uid1",
                req_section="change note 1 req.\nWith multiple lines.",
                opt_section="change note 1 opt.",
                pull_requests=[
                    PullRequest(
                        uid="pr1", closes_threads=("thread1", "thread2"), author_uids=("author1",)
                    ),
                    PullRequest(uid="pr2", author_uids=("author2",)),
                ],
            )
        )
        section_version_note.add_change_note(
            DummySectionChangeNote(slug="slug2", uid="uid2", req_section="change note 2 req.")
        )
        section_version_note.add_change_note(
            DummySectionChangeNote(
                slug="slug3",
                uid="uid3",
                req_section="change note 3 req.",
                pull_requests=[PullRequest(uid="pr_a", author_uids=("author_b", "author_c"))],
            )
        )
        assert (
            section_version_note.render("rst")
            == """\
Optional Section
----------------

- change note 1 opt.

Required Section
----------------

- change note 1 req.

  With multiple lines.

  (`#pr1 <https://github.com/my-username/my-repo/pull/pr1>`_ by `@author1 \
<https://github.com/author1>`_ closes `#thread1 \
<https://github.com/my-username/my-repo/issues/thread1>`_, `#thread2 \
<https://github.com/my-username/my-repo/issues/thread2>`_; `#pr2 \
<https://github.com/my-username/my-repo/pull/pr2>`_ by `@author2 <https://github.com/author2>`_)
- change note 2 req.
- change note 3 req. (`#pr_a <https://github.com/my-username/my-repo/pull/pr_a>`_ by \
`@author_b <https://github.com/author_b>`_, `@author_c <https://github.com/author_c>`_)\
"""
        )



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import abc
from collections.abc import Collection, Iterator

from .._changenoteinfo import ChangeNoteInfo
from .._utils.types import VUIDInput
from .._version import Version
from ..error import ChanGoError


class VersionScanner(Collection[Version]):
    """Abstract base class for a version scanner that can list available versions.

    Hint:
        Objects of this class can be used as :class:`~collections.abc.Collection` of versions
        as returned by the :meth:`get_available_versions` method.
    """

    def __contains__(self, x: object, /) -> bool:
        if x is None:
            return self.has_unreleased_changes()
        if not isinstance(x, str | Version):
            return False
        return self.is_available(x)

    def __iter__(self) -> Iterator[Version]:
        return iter(self.get_available_versions())

    def __len__(self) -> int:
        return len(self.get_available_versions())

    @abc.abstractmethod
    def is_available(self, uid: VUIDInput) -> bool:
        """Check if the version with the given identifier is available.

        Tip:
            :obj:`None` may be passed for convenience, but it's recommended that an implementation
            calls :meth:`has_unreleased_changes` internally.

        Args:
            uid (:class:`~chango.Version` | :obj:`str` | :obj:`None`): The version identifier to
                check.
        """

    @abc.abstractmethod
    def has_unreleased_changes(self) -> bool:
        """Check if there are changes in the repository that are not yet released in a version.

        Returns:
            :obj:`bool`: :obj:`True` if there are unreleased changes, :obj:`False` otherwise.
        """

    @abc.abstractmethod
    def get_latest_version(self) -> Version:
        """Get the latest version

        Returns:
            :class:`~chango.Version`: The latest version

        Raises:
            ~chango.error.ChanGoError: If no versions are available.
        """

    @abc.abstractmethod
    def get_available_versions(
        self, start_from: VUIDInput = None, end_at: VUIDInput = None
    ) -> tuple[Version, ...]:
        """Get the available versions.

        Important:
            Unreleased changes must *not* be included in the returned version identifiers.

        Args:
            start_from (:class:`~chango.Version` | :obj:`str`, optional): The version identifier
                to start from. If :obj:`None`, start from the earliest available version.
            end_at (:class:`~chango.Version` | :obj:`str`, optional): The version identifier to end
                at. If :obj:`None`, end at the latest available version, *excluding* unreleased
                changes.

        Returns:
            Tuple[:class:`~chango.Version`]: The available versions.
        """

    @abc.abstractmethod
    def lookup_change_note(self, uid: str) -> ChangeNoteInfo:
        """Lookup a change note with the given identifier.

        Args:
            uid (:obj:`str`): The unique identifier or file name of the change note to lookup

        Returns:
            :class:`chango.ChangeNoteInfo`: The metadata about the change note specifying the file
                path and version it belongs to.

        Raises:
            ~chango.error.ChanGoError: If the change note with the given identifier is not
                available.
        """

    def get_version(self, uid: str) -> Version:
        """Get the version with the given identifier.

        Hint:
            The default implementation calls :meth:`get_available_versions`. Implementations may
            override this method to provide a more efficient way to get the version.

        Args:
            uid (:obj:`str`): The version identifier to get the version for.

        Returns:
            :class:`~chango.Version`: The version.

        Raises:
            ~chango.error.ChanGoError: If the version with the given identifier is not available.
        """
        try:
            return next(version for version in self if version.uid == uid)
        except StopIteration as exc:
            raise ChanGoError(f"Version '{uid}' not available.") from exc

    @abc.abstractmethod
    def get_changes(self, uid: VUIDInput) -> tuple[str, ...]:
        """Get the changes either for a given version identifier or all available.

        Hint:
            To easily extract the UIDs from the change files,
            :meth:`chango.helpers.change_uid_from_file` can be used.

        Important:
            The returned UIDs must be in the order in which the changes were made.

        Args:
            uid (:class:`~chango.Version` | :obj:`str` | :obj:`None`): The version identifier to
                get the change files for. If :obj:`None`, get the change files for unreleased
                changes must be returned.

        Returns:
            Tuple[:obj:`str`]: UIDs of the changes corresponding to the version identifier.

        Raises:
            ~chango.error.ChanGoError: If the version with the given identifier is not available.
        """

    def invalidate_caches(self) -> None:
        """Invalidate any internal caches that may be used by the implementation.

        Important:

            * This method is not required to do anything if the implementation does not use any
              caches. By default, it does nothing.
            * This method is called by :meth:`chango.abc.ChanGo.release` and
              :meth:`chango.abc.ChanGo.write_change_note` after the respective operation has been
              completed. This gives the implementation the opportunity to clear any caches that
              may have been affected by the operation.
        """



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import abc
from pathlib import Path
from typing import Any, Self

from .._utils.filename import FileName
from .._utils.files import UTF8
from .._utils.types import PathLike
from ..action import ChanGoActionData


class ChangeNote(abc.ABC):
    """Abstract base class for a change note describing a single change in a software project.

    Args:
        slug (:obj:`str`): A short, human-readable identifier for the change note.
        uid (:obj:`str`): A unique identifier for the change note. If not provided, a
            random identifier will be generated. Should be 8 characters long and consist of
            lowercase letters and digits.
    """

    def __init__(self, slug: str, uid: str | None = None):
        self._file_name = FileName(slug=slug, uid=uid) if uid else FileName(slug=slug)

    @property
    def slug(self) -> str:
        """:obj:`str`: The short, human-readable identifier for the change note."""
        return self._file_name.slug

    @property
    def uid(self) -> str:
        """:obj:`str`: The unique identifier for the change note."""
        return self._file_name.uid

    @property
    @abc.abstractmethod
    def file_extension(self) -> str:
        """:obj:`str`: The file extension to use when writing the change note to a file. The
        extension must *not* include the leading dot.
        """

    def update_uid(self, uid: str) -> None:
        """Update the UID of the change note.
        Use with caution.

        Args:
            uid (:obj:`str`): The new UID to use.
        """
        self._file_name = FileName(slug=self.slug, uid=uid)

    @classmethod
    @abc.abstractmethod
    def build_template(cls, slug: str, uid: str | None = None) -> Self:
        """Build a template change note for the concrete change note type.

        Tip:
            This will be used to create a new change note in the CLI.

        Args:
            slug (:obj:`str`): The slug to use for the change note.
            uid (:obj:`str`): The unique identifier for the change note or :obj:`None` to generate
                a random one.

        Returns:
            The :class:`ChangeNote` object.
        """

    @classmethod
    def build_from_github_event(
        cls, event: dict[str, Any], data: dict[str, Any] | ChanGoActionData | None = None
    ) -> Self:
        """Build a change note from a GitHub event.

        Important:
            This is an optional method and by default raises a :class:`NotImplementedError`.
            Implement this method if you want to automatically create change notes based on
            GitHub events.

        Tip:
            This method is useful for automatically creating change note drafts in GitHub actions
            to ensure that each pull request has documented changes.

            .. seealso:: :ref:`action`

        Args:
            event (Dict[:obj:`str`, :obj:`~typing.Any`]): The GitHub event data. This should be one
              of the `events that trigger workflows <ettw>`_. The event is represented as a
              JSON dictionary.
            data (Dict[:obj:`str`, :obj:`~typing.Any`] | :class:`chango.action.ChanGoActionData`, \
               optional): Additional data that may be required to build the change note.

        Returns:
            :class:`CNT <typing.TypeVar>`: The change note or :obj:`None`.

        Raises:
            NotImplementedError: If the method is not implemented.

        .. _ettw: https://docs.github.com/en/actions/writing-workflows/\
            choosing-when-your-workflow-runs/events-that-trigger-workflows
        """
        raise NotImplementedError

    @property
    def file_name(self) -> str:
        """The file name to use when writing the change note to a file."""
        return self._file_name.to_string(self.file_extension)

    @classmethod
    def from_file(cls, file_path: PathLike, encoding: str = UTF8) -> Self:
        """
        Read a change note from the specified file.

        Tip:
            This convenience method calls :meth:`from_bytes` internally.

        Args:
            file_path (:class:`pathlib.Path` | :obj:`str`): The path to the file to read from.
            encoding (:obj:`str`): The encoding to use for reading.

        Returns:
            :class:`ChangeNote`: The :class:`ChangeNote` object.

        Raises:
            :class:`chango.error.ValidationError`: If the data is not a valid change note file.
        """
        path = Path(file_path)
        file_name = FileName.from_string(path.name)
        return cls.from_bytes(
            slug=file_name.slug, uid=file_name.uid, data=path.read_bytes(), encoding=encoding
        )

    @classmethod
    def from_bytes(cls, slug: str, uid: str, data: bytes, encoding: str = UTF8) -> Self:
        """
        Read a change note from the specified byte data. The data will be the raw binary content
        of a change note file.

        Tip:
            This convenience method calls :meth:`from_string` internally.

        Args:
            slug (:obj:`str`): The slug of the change note.
            uid (:obj:`str`): The UID of the change note.
            data (:obj:`bytes`): The bytes to read from.
            encoding (:obj:`str`): The encoding to use for reading.

        Returns:
            :class:`ChangeNote`: The :class:`ChangeNote` object.

        Raises:
            :class:`chango.error.ValidationError`: If the data is not a valid change note file.
        """
        return cls.from_string(slug=slug, uid=uid, string=data.decode(encoding))

    @classmethod
    @abc.abstractmethod
    def from_string(cls, slug: str, uid: str, string: str) -> Self:
        """Read a change note from the specified string data. The implementation must be able to
        handle the case where the string is not a valid change note and raise an
        :exc:`~chango.error.ValidationError` in that case.

        Args:
            slug (:obj:`str`): The slug of the change note.
            uid (:obj:`str`): The UID of the change note.
            string (:obj:`str`): The string to read from.

        Returns:
            :class:`ChangeNote`: The :class:`ChangeNote` object.

        Raises:
            :class:`chango.error.ValidationError`: If the string is not a valid change note.
        """

    def to_bytes(self, encoding: str = UTF8) -> bytes:
        """Write the change note to bytes. This binary data should be suitable for writing to a
        file and reading back in with :meth:`from_bytes`.

        Tip:
            This convenience method calls :meth:`to_string` internally.

        Args:
            encoding (:obj:`str`): The encoding to use.

        Returns:
            :obj:`bytes`: The bytes data.
        """
        return self.to_string(encoding).encode(encoding)

    @abc.abstractmethod
    def to_string(self, encoding: str = UTF8) -> str:
        """Write the change note to a string. This string should be suitable for writing to a file
        and reading back in with :meth:`from_string`.

        Args:
            encoding (:obj:`str`): The encoding to use for writing.

        Returns:
            :obj:`str`: The string data.
        """

    def to_file(self, directory: PathLike | None = None, encoding: str = UTF8) -> Path:
        """Write the change note to the directory.

        Hint:
            The file name will always be the :attr:`~chango.abc.ChangeNote.file_name`.

        Args:
            directory: Optional. The directory to write the file to. If not provided, the file
                will be written to the current working directory.
            encoding (:obj:`str`): The encoding to use for writing.

        Returns:
            :class:`pathlib.Path`: The path to the file that was written.
        """
        path = Path(directory) if directory else Path.cwd()
        write_path = path / self.file_name
        write_path.write_bytes(self.to_bytes(encoding=encoding))
        return write_path
